import dataclasses
from enum import Enum

from .configurations import ActivationCheckpointingType
from .data import Size, TensorRepr
from .math import safe_divide


@dataclasses.dataclass
class ParallelConfig:
    class ZeroLevel(int, Enum):
        NONE = 0
        PARTITION_OPTIMIZER = 1
        PARTITION_GRADIENTS = 2
        PARTITION_PARAMETERS = 3

    @dataclasses.dataclass
    class ExpertParallelCfg:
        ep: int  # Expert Parallel (EP) degree
        tp: int
        dp: int

    tp: int  # Tensor Parallel (TP) degree
    cp: int  # Context Parallel (CP) degree
    pp: int  # Pipeline Parallel (PP) degree
    dp: int  # Data Parallel (DP) degree

    expert_mesh: ExpertParallelCfg | None

    vpp: int  # Virtual Pipeline Parallel (VPP) degree

    sp_enabled: bool  # Sequence Parallel (SP) enablement

    zero_level: ZeroLevel

    def __post_init__(self) -> None:
        if self.expert_mesh is not None:
            assert (
                self.expert_mesh.ep * self.expert_mesh.tp * self.expert_mesh.dp
                == self.dp * self.cp * self.tp
            )

    def world_size(self) -> int:
        return self.tp * self.cp * self.pp * self.dp


class DistributedAdamOptimizerStates:
    """Optimizer states from Apex DistributedFusedAdam.
    see: https://github.com/NVIDIA/Megatron-LM/blob/main/docs/source/distrib_optimizer.md
    and: https://github.com/NVIDIA/apex/blob/master/apex/contrib/optimizers/distributed_fused_adam.py

    the distributed optimizer recieves a set of parameters to manage that are already
    model parallel partitioned. Internally, it will additionally partition states
    over DP, so optimizer states end up being partitioned over MP * DP, but
    the diestributed optimizer doesn't have any concept of model parallelism.

    NOTE: shards will be larger in reality due to alignment requirements and
    unfilled buckets.
    """

    # NOTE: n_params here is meant to be the parameters in a pipeline stage.
    def __init__(self, n_params: int, store_param_remainders: bool, dp: int) -> None:
        self.param_shard = TensorRepr(
            unpartitioned_shape=(n_params,),
            partition_spec={0: dp},
            # apex has a optimization to avoid storing information that's redundant
            # between fp32 and fp16 weights. it can instead store an extra 16
            # bits of precision, which can be combined with bf16 weights to yield
            # fp32 weights
            bits_per_elt=16 if store_param_remainders else 32,
            enforce_evenly_partitionable=False,
        )
        self.exp_avg_shard = TensorRepr(
            unpartitioned_shape=(n_params,),
            partition_spec={0: dp},
            bits_per_elt=32,
            enforce_evenly_partitionable=False,
        )
        self.exp_avg_sq_shard = TensorRepr(
            unpartitioned_shape=(n_params,),
            partition_spec={0: dp},
            bits_per_elt=32,
            enforce_evenly_partitionable=False,
        )
        # reused for various purposes but at one point it needs to hold
        # all (i.e. not DP partitioned) parameter gradients, which are accumulated
        # to until they are reduce-scattered to a gradient shard during the final
        # microbatch.
        # basically ZeRO2, but with no reduce-scatter between gradient accumulations.
        self.grad_buffer = TensorRepr(
            unpartitioned_shape=(n_params,),
            partition_spec={},
            bits_per_elt=32,
        )

    def total_bytes(self, partitioned: bool) -> int:
        return _sum(
            self.param_shard.size(partitioned=partitioned).bytes(),
            self.exp_avg_shard.size(partitioned=partitioned).bytes(),
            self.exp_avg_sq_shard.size(partitioned=partitioned).bytes(),
            self.grad_buffer.size(partitioned=partitioned).bytes(),
        )


@dataclasses.dataclass
class ModelStates:
    """Tracks persistent tensor allocations kept throughput training."""

    params_shard: TensorRepr
    opt_states: DistributedAdamOptimizerStates

    def total_bytes(self, partitioned: bool) -> int:
        return self.params_shard.size(
            partitioned=partitioned
        ).bytes() + self.opt_states.total_bytes(partitioned=partitioned)

    def __repr__(self) -> str:
        params_bytes = self.params_shard.size(partitioned=True).bytes()
        opt_states_bytes = self.opt_states.total_bytes(partitioned=True)

        return "\n".join(
            [
                f"params          : {self.params_shard.size(partitioned=True)}",
                f"params (opt)    : {self.opt_states.param_shard.size(partitioned=True)}",
                f"exp_avg         : {self.opt_states.exp_avg_shard.size(partitioned=True)}",
                f"exp_avg_squared : {self.opt_states.exp_avg_sq_shard.size(partitioned=True)}",
                f"grad_buffer     : {self.opt_states.grad_buffer.size(partitioned=True)}",
                f"TOTAL           : {(params_bytes + opt_states_bytes) / (1024**3):.2f}GiB",
            ]
        )


def _sum(*summands: int) -> int:
    return sum(list(summands))


@dataclasses.dataclass
class MoeCfg:
    n_experts: int
    expert_inter_sz: int
    experts_per_token: int
    capacity_factor: float
    moe_frequency: float
    expert_tp_degree: int


@dataclasses.dataclass
class ThreeDParallelModel:
    """Representation of a 3D parallel transformer model."""

    parallelism_cfg: ParallelConfig

    # Instance variables set in __post_init__
    mlp_up_exp_weight: TensorRepr | None = dataclasses.field(init=False, default=None)
    mlp_down_exp_weight: TensorRepr | None = dataclasses.field(init=False, default=None)

    sequence_len: int
    microbatch_sz: int

    hidden_sz: int

    n_layers: int

    n_q_heads: int
    n_kv_heads: int  # num_query_groups if GQA
    head_dim: int

    inter_sz: int
    glu: bool
    moe_cfg: MoeCfg | None

    rotary_embed: bool

    dropout: bool

    vocab_sz: int
    tie_embeddings: bool

    act_ckpting_type: ActivationCheckpointingType

    n_param_buckets: int

    # TODO. assuming mixed precision here.
    bits_per_parameter: int = 16
    bits_per_grad: int = 32
    bits_per_optim_state: int = 32

    def __post_init__(self) -> None:
        if self.n_layers % (self.parallelism_cfg.pp * self.parallelism_cfg.vpp) != 0:
            raise ValueError(
                f"number of layers {self.n_layers} is not divisible by the product "
                f"of PP={self.parallelism_cfg.pp} and VPP={self.parallelism_cfg.vpp}"
            )

        if self.moe_cfg:
            if not (self.moe_cfg.moe_frequency * self.n_layers).is_integer():
                raise ValueError(
                    f"invalid moe frequency {self.moe_cfg.moe_frequency} for layer number {self.n_layers}"
                )
            self.n_moe_layers = int(self.moe_cfg.moe_frequency * self.n_layers)
            self.n_nml_layers = self.n_layers - self.n_moe_layers
        else:
            self.n_moe_layers = 0
            self.n_nml_layers = self.n_layers

        n_experts = self.moe_cfg.n_experts if self.moe_cfg else 1

        self.embed_weight = TensorRepr(
            unpartitioned_shape=(self.hidden_sz, self.vocab_sz),
            partition_spec={1: self.parallelism_cfg.tp},  # vocab-parallel
            bits_per_elt=self.bits_per_parameter,
        )
        self.pre_attn_norm_weight = TensorRepr(
            unpartitioned_shape=(self.hidden_sz,),
            partition_spec={},  # replicated
            bits_per_elt=self.bits_per_parameter,
        )
        self.qkv_weight = TensorRepr(
            unpartitioned_shape=(
                self.hidden_sz,
                # following common practice of merging Q + K + V matmuls.
                (self.n_q_heads + 2 * self.n_kv_heads) * self.head_dim,
            ),
            partition_spec={1: self.parallelism_cfg.tp},  # col parallel
            bits_per_elt=self.bits_per_parameter,
        )
        self.attn_out_weight = TensorRepr(
            unpartitioned_shape=(self.hidden_sz, self.hidden_sz),
            partition_spec={0: self.parallelism_cfg.tp},  # row parallel
            bits_per_elt=self.bits_per_parameter,
        )
        self.router_weight = TensorRepr(
            unpartitioned_shape=(
                self.hidden_sz,
                self.moe_cfg.n_experts if self.moe_cfg else 0,
            ),
            partition_spec={},
            bits_per_elt=self.bits_per_parameter,
        )
        self.mlp_up_weight = TensorRepr(
            unpartitioned_shape=(
                self.hidden_sz,
                # following common practice of merging up + gate matmuls in the event
                # we're using GLU.
                (self.inter_sz * 2) if self.glu else self.inter_sz,
            ),
            partition_spec={1: self.parallelism_cfg.tp},  # col parallel
            bits_per_elt=self.bits_per_parameter,
        )
        self.mlp_down_weight = TensorRepr(
            unpartitioned_shape=(self.inter_sz, self.hidden_sz),
            partition_spec={0: self.parallelism_cfg.tp},  # row parallel
            bits_per_elt=self.bits_per_parameter,
        )

        if self.moe_cfg is not None:
            assert self.parallelism_cfg.expert_mesh is not None
            self.mlp_up_exp_weight = (
                TensorRepr(
                    # following common practice of merging up + gate matmuls in the event
                    # we're using GLU.
                    unpartitioned_shape=(
                        n_experts,
                        self.hidden_sz,
                        (self.moe_cfg.expert_inter_sz * 2)
                        if self.glu
                        else self.moe_cfg.expert_inter_sz,
                    ),
                    partition_spec={
                        0: self.parallelism_cfg.expert_mesh.ep,
                        2: self.parallelism_cfg.expert_mesh.tp,
                    },  # col parallel
                    bits_per_elt=self.bits_per_parameter,
                )
                if self.moe_cfg is not None
                else None
            )
            self.mlp_down_exp_weight = (
                TensorRepr(
                    unpartitioned_shape=(
                        n_experts,
                        self.moe_cfg.expert_inter_sz,
                        self.hidden_sz,
                    ),
                    partition_spec={
                        0: self.parallelism_cfg.expert_mesh.ep,
                        1: self.parallelism_cfg.expert_mesh.tp,
                    },  # row parallel
                    bits_per_elt=self.bits_per_parameter,
                )
                if self.moe_cfg is not None
                else None
            )
        else:
            self.mlp_up_exp_weight = None
            self.mlp_down_exp_weight = None

        self.pre_mlp_norm_weight = TensorRepr(
            unpartitioned_shape=(self.hidden_sz,),
            partition_spec={},  # replicated
            bits_per_elt=self.bits_per_parameter,
        )

        if self.parallelism_cfg.zero_level != ParallelConfig.ZeroLevel.PARTITION_OPTIMIZER:
            raise NotImplementedError

        self.states = ModelStates(
            params_shard=TensorRepr(
                unpartitioned_shape=(
                    self.__get_n_total_params(
                        spmd_partitioned=True,
                        mpmd_partitioned=True,
                    ),
                ),
                partition_spec={},
                bits_per_elt=self.bits_per_parameter,
                enforce_evenly_partitionable=False,
            ),
            opt_states=DistributedAdamOptimizerStates(
                n_params=self.__get_n_total_params(
                    spmd_partitioned=True,
                    mpmd_partitioned=True,
                ),
                store_param_remainders=True,  # TODO. should be configurable
                dp=self.parallelism_cfg.dp,
            ),
        )

    def get_single_microbatch_fwd_flops(self) -> float:
        return (
            2  # FLOPs/MAC
            * 1  # factor for forward only (1 GEMMs per op)
            * self.microbatch_sz
            * self.sequence_len
            * self.__get_n_active_params(partitioned=False)
        )

    def get_single_microbatch_bwd_flops(self) -> float:
        return (
            2  # FLOPs/MAC
            * 2  # factor for backward only (2 GEMMs per op)
            * self.microbatch_sz
            * self.sequence_len
            * self.__get_n_active_params(partitioned=False)
        )

    def get_n_total_params(self, partitioned: bool) -> int:
        return self.__get_n_total_params(
            spmd_partitioned=partitioned,
            mpmd_partitioned=partitioned,
        )

    def get_n_active_params(self, partitioned: bool) -> int:
        return self.__get_n_active_params(partitioned=partitioned)

    def activation_size_per_microbatch_per_layer(self) -> Size:
        return Size(
            numel=self.__dense_activation_numel_per_microbatch_per_layer(),
            bits_per_element=self.bits_per_parameter,
        )

    def vpp_penalty(self) -> float:
        """interleaved schedule requires storing activations for (1 + (p - 1)/pm)
        more layers."""
        if self.parallelism_cfg.vpp == 1:
            return 1.0

        return 1 + (self.parallelism_cfg.pp - 1) / (
            self.parallelism_cfg.pp * self.parallelism_cfg.vpp
        )

    def layers_per_pp_stage(self) -> int:
        return sum(self.__n_layers(mpmd_partitioned=True, moe=moe) for moe in [False, True])

    def __get_n_total_params(self, spmd_partitioned: bool, mpmd_partitioned: bool) -> int:
        return _sum(
            # we'll give the number of parameters on the most heavily loaded pipeline stage
            # if PP=1 then the only pipeline stage must store both embedding and LM head.
            (1 if (mpmd_partitioned and self.parallelism_cfg.pp > 1) or self.tie_embeddings else 2)
            * self.__get_embedding_or_lm_head_size(spmd_partitioned=spmd_partitioned),
            # add in the transformer blocks
            sum(
                self.__n_layers(mpmd_partitioned=mpmd_partitioned, moe=moe)
                * self.__get_transformer_block_n_params(
                    spmd_partitioned=spmd_partitioned,
                    moe=moe,
                    active=False,
                    experts_per_token=None,
                )
                # TODO. clean this up
                for moe in ([False, True] if self.moe_cfg is not None else [False])
            ),
        )

    def __get_n_active_params(self, partitioned: bool) -> int:
        if not self.moe_cfg:
            return self.__get_n_total_params(
                spmd_partitioned=partitioned,
                mpmd_partitioned=partitioned,
            )

        # we'll give the number of parameters on the most heavily loaded pipeline stage
        # if PP=1 then the only pipeline stage must store both embedding and LM head.
        return _sum(
            # embedding/lmhead
            (1 if (partitioned and self.parallelism_cfg.pp > 1) or self.tie_embeddings else 2)
            * self.__get_embedding_or_lm_head_size(spmd_partitioned=partitioned),
            # transformer blocks
            sum(
                self.__n_layers(mpmd_partitioned=partitioned, moe=moe)
                * self.__get_transformer_block_n_params(
                    spmd_partitioned=partitioned,
                    moe=moe,
                    active=True,
                    experts_per_token=self.moe_cfg.experts_per_token,
                )
                for moe in [False, True]
            ),
        )

    def __get_embedding_or_lm_head_size(self, spmd_partitioned: bool) -> int:
        return self.embed_weight.size(partitioned=spmd_partitioned).numel()

    def __get_transformer_block_n_params(
        self,
        spmd_partitioned: bool,
        moe: bool,
        # TODO. would rather not expose these.
        active: bool,
        experts_per_token: int | None,
    ) -> int:
        assert active == (experts_per_token is not None)

        if moe:
            assert self.moe_cfg is not None
            assert self.mlp_up_exp_weight is not None
            assert self.mlp_down_exp_weight is not None
            mlp_params = (
                # if we're trying to compute active params, then we account for the
                # fact that we'll apply topk mlps per token.
                experts_per_token  # type: ignore[operator]
                * _sum(
                    self.mlp_up_exp_weight.numel(partitioned=spmd_partitioned)
                    // self.moe_cfg.n_experts,
                    self.mlp_down_exp_weight.numel(partitioned=spmd_partitioned)
                    // self.moe_cfg.n_experts,
                )
                if active
                else _sum(
                    self.mlp_up_exp_weight.numel(partitioned=spmd_partitioned),
                    self.mlp_down_exp_weight.numel(partitioned=spmd_partitioned),
                )
            )
        else:
            mlp_params = _sum(
                self.mlp_up_weight.numel(partitioned=spmd_partitioned),
                self.mlp_down_weight.numel(partitioned=spmd_partitioned),
            )

        return _sum(
            self.pre_attn_norm_weight.numel(partitioned=spmd_partitioned),
            self.qkv_weight.numel(partitioned=spmd_partitioned),
            self.attn_out_weight.numel(partitioned=spmd_partitioned),
            self.pre_mlp_norm_weight.numel(partitioned=spmd_partitioned),
            mlp_params,
        )

    def __n_layers(self, mpmd_partitioned: bool, moe: bool) -> int:
        # TODO. we're making the assumption that MoE and non-MoE layers
        # can be evenly partitioned.
        n_layers = self.n_moe_layers if moe else self.n_nml_layers
        if mpmd_partitioned:
            n_layers = safe_divide(n_layers, self.parallelism_cfg.pp)

        return n_layers

    # TODO. probably needs to be adjusted for MoE
    def __dense_activation_numel_per_microbatch_per_layer(self) -> int:
        """
        See: Reducing Activation Recomputation in Large Transformer Models
        https://arxiv.org/pdf/2205.05198.pdf
        """
        sbh = self.sequence_len * self.microbatch_sz * self.hidden_sz
        sbq = self.sequence_len * self.microbatch_sz * self.n_q_heads * self.head_dim
        sbk = self.sequence_len * self.microbatch_sz * self.n_kv_heads * self.head_dim
        sbv = self.sequence_len * self.microbatch_sz * self.n_kv_heads * self.head_dim

        n_local_mlps = (
            1
            if self.moe_cfg is None or self.parallelism_cfg.expert_mesh is None
            else safe_divide(self.moe_cfg.n_experts, self.parallelism_cfg.expert_mesh.ep)
        )
        sbi = (
            self.sequence_len * self.microbatch_sz * self.inter_sz
            if self.moe_cfg is None
            else self.expert_capacity() * self.moe_cfg.expert_inter_sz
        )

        if self.act_ckpting_type == ActivationCheckpointingType.FULL:
            return self.__sp_partition_if_on(sbh)  # just the block input
        elif self.act_ckpting_type == ActivationCheckpointingType.SUPER_SELECTIVE:
            return _sum(
                # LAYERNORM 1
                # - output is recomputed
                # QKV (col parallel linear)
                self.__tp_partition(sbq),  # Q - attn input
                self.__tp_partition(sbk),  # K - attn input
                self.__tp_partition(sbv),  # V - attn input
                # ROTARY EMBEDDINGS
                self.__tp_partition(sbq) if self.rotary_embed else 0,  # Q rotary
                self.__tp_partition(sbk) if self.rotary_embed else 0,  # K rotary
                # SELF ATTENTION
                # - skipping intermediates (checkpointed by FlashAttention)
                self.__tp_partition(sbh),  # attn output
                # DOWN PROJ
                # - output deallocated (dropout doesn't need to store)
                # DROPOUT
                # - dropout mask recomputed
                # - deallocated - residual doesn't need to store
                # RESIDUAL
                self.__sp_partition_if_on(sbh),  # stored by norm2, residual2
                # LAYERNORM 2
                self.__sp_partition_if_on(sbh),
                # TOKEN PERMUTATION (ALLGATHERED)
                # TODO.
                # output is recomputed
                # MLP UP/GATE (col parallel linear)
                n_local_mlps * self.__tp_partition(2 * sbi if self.glu else sbi),  # SwiGLU input
                # SwiGLU
                # TODO. swiglu output
                # MLP DOWN (row parallel linear)
                # - output deallocated - dropout doesn't need to store
                # TODO. all-to-all token unpermutation
                # TOKEN UNPERMUTATION
                # - output is recomputed
                # DROPOUT
                # - dropout mask recomputed
                # - output deallocated - residual doesn't need to store
                # RESIDUAL
                self.__sp_partition_if_on(sbh),  # needed by next norm1, residual1
            )
        elif self.act_ckpting_type == ActivationCheckpointingType.SELECTIVE:
            return _sum(
                # LAYERNORM 1
                self.__sp_partition_if_on(sbh),  # output - QKV input
                # QKV PROJ (col parallel linear)
                self.__tp_partition(sbq),  # Q - attn input
                self.__tp_partition(sbk),  # K - attn input
                self.__tp_partition(sbv),  # V - attn input
                # ROTARY EMBEDDINGS
                self.__tp_partition(sbq) if self.rotary_embed else 0,  # Q rotary
                self.__tp_partition(sbk) if self.rotary_embed else 0,  # K rotary
                # SELF ATTENTION
                # - skipping intermediates (checkpointed by FlashAttention)
                self.__tp_partition(sbh),  # needed by down proj
                # DOWN PROJ
                # - output deallocated (dropout doesn't need to store)
                # DROPOUT
                self.__sp_partition_if_on(int(0.5 * sbh)) if self.dropout else 0,  # dropout mask
                # -  output deallocated: residual doesn't need to store
                # RESIDUAL
                self.__sp_partition_if_on(sbh),  # needed by norm2, resid2
                # LAYERNORM 2
                self.__sp_partition_if_on(sbh),  # up/gate input
                # MLP UP/GATE (col parallel linear)
                n_local_mlps * self.__tp_partition(2 * sbi if self.glu else sbi),  # SwiGLU input
                # SwiGLU
                self.__tp_partition(sbi),  # SiLU output
                self.__tp_partition(sbi),  # gate output
                # MLP DOWN (row parallel linear)
                # - output deallocated - dropout doesn't need to store
                # DROPOUT
                self.__sp_partition_if_on(int(0.5 * sbh)) if self.dropout else 0,  # dropout mask
                # - output deallocated - residual doesn't need to store
                # RESIDUAL
                # input to next norm1, resid1
                self.__sp_partition_if_on(sbh),
            )
        elif self.act_ckpting_type == ActivationCheckpointingType.NONE:
            raise NotImplementedError(
                "not yet implemented. selective will give a good approximation "
                "if you use flash attention"
            )
        else:
            raise ValueError(f"unhandled checkpointing_type={self.act_ckpting_type}")

    def __tp_partition(self, unpartitoned_numel: int) -> int:
        x = unpartitoned_numel
        for parallelism_degree in [self.parallelism_cfg.cp, self.parallelism_cfg.tp]:
            x = safe_divide(x, parallelism_degree)

        return x

    def __sp_partition_if_on(self, unpartitioned_numel: int) -> int:
        parallelism_degrees = [self.parallelism_cfg.cp]
        if self.parallelism_cfg.sp_enabled:
            parallelism_degrees.append(self.parallelism_cfg.tp)

        x = unpartitioned_numel
        for parallelism_degree in parallelism_degrees:
            x = safe_divide(x, parallelism_degree)

        return x

    def expert_capacity(self) -> int:
        """Returns the number of tokens that can be processed by each expert."""
        if self.moe_cfg is None:
            raise RuntimeError
        assert self.parallelism_cfg.expert_mesh is not None

        return safe_divide(
            int(
                safe_divide(self.sequence_len, self.parallelism_cfg.tp * self.parallelism_cfg.cp)
                * self.microbatch_sz
                * self.moe_cfg.experts_per_token
                * self.parallelism_cfg.expert_mesh.ep
                * self.moe_cfg.capacity_factor
            ),
            self.moe_cfg.n_experts,
        )



"""
given a kubernetes pod name prefix for some compute cluster, retrieve AWS network
topology info and plot.
"""

import re
from argparse import ArgumentParser

import boto3  # type: ignore[import-untyped, import-not-found]
from pyvis.network import Network

import dlcalc.utils.cluster.ec2
import dlcalc.utils.cluster.kubernetes

pod_id_pattern = re.compile(r"worker-(\d+)")


def _dict_to_label(dict: dict) -> str:
    kv_pairs = []
    for key, value in dict.items():
        kv_pairs.append(f"{key}: {value}")
    return "\n".join(kv_pairs)


def main() -> None:
    parser = ArgumentParser(__doc__)
    parser.add_argument(
        "search_prefix",
        help="kubernetes pod name prefix to use when filtering pods returned by `kubectl get pods`",
    )
    args = parser.parse_args()

    job_search_prefix = args.search_prefix

    ######################################################################################
    # STEP 1: Get the list of instance IDs involved in a kubernetes job
    ######################################################################################
    cluster_members = dlcalc.utils.cluster.kubernetes.get_kubernetes_cluster_members(
        job_search_prefix
    )

    ######################################################################################
    # STEP 2: Get topology of the instances acquired earlier.
    ######################################################################################
    instance_ids = {m.node_instance_id for m in cluster_members}
    (cluster_region,) = set(m.node_region for m in cluster_members)
    cluster_instance_types = {m.node_instance_type for m in cluster_members}
    cluster_azs = {m.node_az for m in cluster_members}
    instance_id_to_job_member = {m.node_instance_id: m for m in cluster_members}

    print(f"getting topology info for instance IDs {instance_ids}")
    ec2_client = boto3.client("ec2", region_name=cluster_region)

    net = Network(directed=True)
    for instance_info in dlcalc.utils.cluster.ec2.iter_instance_info(
        ec2_client=ec2_client,
        accepted_node_instance_types=cluster_instance_types,
        accepted_node_availability_zones=cluster_azs,
        accepted_instance_ids=instance_ids,
    ):
        job_member = instance_id_to_job_member[instance_info.instance_id]

        net.add_node(
            instance_info.instance_id,
            title=_dict_to_label(
                {
                    "PodName": job_member.pod_name,
                    "WorkerId": job_member.worker_id,
                    "InstanceType": instance_info.instance_type,
                    "AvailabilityZone": instance_info.instance_az,
                }
            ),
            color="blue",
            physics=False,
        )

        for network_node in instance_info.network_nodes:
            net.add_node(network_node, color="red", physics=False)

        # order of network nodes: farthest from instance -> closest to instance
        chain = [*instance_info.network_nodes, instance_info.instance_id]
        for edge_src_idx in range(len(chain) - 1):
            edge_dst_idx = edge_src_idx + 1
            net.add_edge(chain[edge_src_idx], chain[edge_dst_idx], physics=False)

    net.show_buttons()
    net.show("graph.html", notebook=False)


if __name__ == "__main__":
    main()



"""
Reduce-Scatter:
* Ring: https://github.com/NVIDIA/nccl/blob/ab2b89c4c339bd7f816fbc114a4b05d386b66290/src/device/reduce_scatter.h#L12-L54

All-Gather:
* Ring: https://github.com/NVIDIA/nccl/blob/ab2b89c4c339bd7f816fbc114a4b05d386b66290/src/device/all_gather.h#L12-L64

All-Reduce (not relevant for comm patterns we have represented for now):
some info on tree AR: https://github.com/NVIDIA/nccl/issues/545

TODO. try to get estimates closer by accounting for NCCL protocols https://github.com/NVIDIA/nccl/issues/281
"""

from enum import Enum

from .configurations import CrossDCConfig
from .data import Size
from .hardware import MachineSpec
from .model_3d import ParallelConfig


class ParallelismType(Enum):
    """Types of parallelism in hierarchical order."""

    PP = 0  # Pipeline Parallel
    DP = 1  # Data Parallel
    CP = 2  # Context Parallel
    TP = 3  # Tensor Parallel
    # Expert parallelism types
    EDP = 1  # Expert Data Parallel (same level as DP)
    EP = 2  # Expert Parallel (same level as CP)
    ETP = 3  # Expert Tensor Parallel (same level as TP)


def _get_effective_bw(
    *,
    parallelism_type: ParallelismType,
    parallel_config: ParallelConfig,
    machine_spec: MachineSpec,
    is_expert_comm: bool,
) -> float:
    """Calculate effective bandwidth for a given parallelism type.

    Few things we need to figure out:

    Which kind of link the communication will utilize.
    -------------------------------------------------------------------------------------------
    Given some non-overlapping parallelism order like [PP, DP, CP, TP] or [PP, eDP, EP, eTP]
    if the product of parallelisms including and after the parallelism in question is less than
    or equal to the number of workers per node, then we will use intra-node links. Otherwise
    we will use inter-node links.

    How many participants will share the link.
    -------------------------------------------------------------------------------------------
    If the products of the parallelisms coming after the parallelism in question is greater than
    or equal to the number of workers per node, then the cross-node bandwidth is divided by the
    number of workers per node. Otherwise, it is divided by the number of workers per node divided
    product of following parallelisms.

    NOTE: we only account for link being shared across one communication type at a time. We do
    not account for cross-parallelism competition (like for example overlapped DP comms at the
    same time as cross-node EP comms).

    Args:
        parallelism_type: The type of parallelism for which to calculate bandwidth
        parallel_config: The parallelism configuration
        machine_spec: The machine specification
        is_expert_comm: Whether this is for expert parallelism communication

    Returns:
        Effective bandwidth in bytes per second
    """
    n_devices_per_node = machine_spec.n_devices

    if is_expert_comm:
        assert parallel_config.expert_mesh is not None
        parallelism_values = {
            ParallelismType.PP: parallel_config.pp,
            ParallelismType.EDP: parallel_config.expert_mesh.dp,
            ParallelismType.EP: parallel_config.expert_mesh.ep,
            ParallelismType.ETP: parallel_config.expert_mesh.tp,
        }
        hierarchy = [
            ParallelismType.PP,
            ParallelismType.EDP,
            ParallelismType.EP,
            ParallelismType.ETP,
        ]
    else:
        parallelism_values = {
            ParallelismType.PP: parallel_config.pp,
            ParallelismType.DP: parallel_config.dp,
            ParallelismType.CP: parallel_config.cp,
            ParallelismType.TP: parallel_config.tp,
        }
        hierarchy = [ParallelismType.PP, ParallelismType.DP, ParallelismType.CP, ParallelismType.TP]

    if parallelism_type not in hierarchy:
        raise AssertionError(
            f"Parallelism type {parallelism_type} not found in hierarchy {hierarchy}."
        )

    product_including_current = parallelism_values[parallelism_type]
    product_after_current = 1
    for parallelism_type in hierarchy[hierarchy.index(parallelism_type) + 1 :]:
        product_including_current *= parallelism_values[parallelism_type]
        product_after_current *= parallelism_values[parallelism_type]

    if product_including_current <= n_devices_per_node:  # use intra-node link (assume no sharing)
        return machine_spec.intra_node_connect.unidirectional_bw_bytes_per_sec

    # otherwise, use inter-node link
    base_bw = machine_spec.inter_node_connect.unidirectional_bw_bytes_per_sec
    # TODO. what if we have something like TP2? there would be 4 TP groups per node.
    n_groups_per_node = n_devices_per_node

    return base_bw / n_groups_per_node


def get_tp_reduce_scatter_comm_time_s(
    size: Size, parallel_config: ParallelConfig, machine_spec: MachineSpec
) -> float:
    """assumes ring algorithm."""
    return _get_ring_tp_ag_or_rs_comm_time_s(
        size,
        n_participants=parallel_config.tp,
        machine_spec=machine_spec,
        parallel_config=parallel_config,
        is_expert_comm=False,
    )


def get_tp_all_gather_comm_time_s(
    size: Size, parallel_config: ParallelConfig, machine_spec: MachineSpec
) -> float:
    """assumes ring algorithm."""
    return _get_ring_tp_ag_or_rs_comm_time_s(
        size,
        n_participants=parallel_config.tp,
        machine_spec=machine_spec,
        parallel_config=parallel_config,
        is_expert_comm=False,
    )


def get_dp_reduce_scatter_latency_term_s(
    parallel_config: ParallelConfig,
    machine_spec: MachineSpec,
) -> float:
    """assumes ring algorithm."""
    return _ring_ag_or_rs_latency_term_s(
        # approximation: we assume most of the parameters are MoE parameters, and calculate
        # just for those. for most models this is good enough.
        n_participants=parallel_config.dp
        if parallel_config.expert_mesh is None
        else parallel_config.expert_mesh.dp,
        link_latency_s=machine_spec.inter_node_connect.latency_sec,
    )


def get_dp_reduce_scatter_bw_term_s(
    size: Size,
    parallel_config: ParallelConfig,
    machine_spec: MachineSpec,
) -> float:
    """assumes ring algorithm."""
    # approximation: we assume most of the parameters are MoE parameters, and calculate
    # just for those. for most models this is good enough.
    return _ring_ag_or_rs_bw_term_s(
        size,
        n_participants=parallel_config.dp
        if parallel_config.expert_mesh is None
        else parallel_config.expert_mesh.dp,
        unidirectional_link_bw_bytes_per_sec=int(
            _get_effective_bw(
                parallelism_type=ParallelismType.DP
                if parallel_config.expert_mesh is None
                else ParallelismType.EDP,
                parallel_config=parallel_config,
                machine_spec=machine_spec,
                is_expert_comm=False,
            )
        ),
    )


def get_dp_reduce_scatter_comm_time_s(
    size: Size,
    parallel_config: ParallelConfig,
    machine_spec: MachineSpec,
) -> float:
    """assumes ring algorithm."""
    latency_term = get_dp_reduce_scatter_latency_term_s(
        parallel_config=parallel_config,
        machine_spec=machine_spec,
    )

    bw_term = get_dp_reduce_scatter_bw_term_s(
        size,
        parallel_config=parallel_config,
        machine_spec=machine_spec,
    )

    return latency_term + bw_term


def get_dp_all_gather_latency_term_s(
    parallel_config: ParallelConfig,
    machine_spec: MachineSpec,
) -> float:
    """assumes ring algorithm."""
    return _ring_ag_or_rs_latency_term_s(
        # approximation: we assume most of the parameters are MoE parameters, and calculate
        # just for those. for most models this is good enough.
        n_participants=parallel_config.dp
        if parallel_config.expert_mesh is None
        else parallel_config.expert_mesh.dp,
        link_latency_s=machine_spec.inter_node_connect.latency_sec,
    )


def get_dp_all_gather_bw_term_s(
    size: Size,
    parallel_config: ParallelConfig,
    machine_spec: MachineSpec,
) -> float:
    """assumes ring algorithm."""
    # approximation: we assume most of the parameters are MoE parameters, and calculate
    # just for those. for most models this is good enough.
    return _ring_ag_or_rs_bw_term_s(
        size,
        n_participants=parallel_config.dp
        if parallel_config.expert_mesh is None
        else parallel_config.expert_mesh.dp,
        unidirectional_link_bw_bytes_per_sec=int(
            _get_effective_bw(
                parallelism_type=ParallelismType.DP
                if parallel_config.expert_mesh is None
                else ParallelismType.EDP,
                parallel_config=parallel_config,
                machine_spec=machine_spec,
                is_expert_comm=False,
            )
        ),
    )


def get_dp_all_gather_comm_time_s(
    size: Size,
    parallel_config: ParallelConfig,
    machine_spec: MachineSpec,
) -> float:
    """assumes ring algorithm."""
    latency_term = get_dp_all_gather_latency_term_s(
        parallel_config=parallel_config,
        machine_spec=machine_spec,
    )

    bw_term = get_dp_all_gather_bw_term_s(
        size,
        parallel_config=parallel_config,
        machine_spec=machine_spec,
    )

    return latency_term + bw_term


def _ring_ag_or_rs_latency_term_s(
    n_participants: int,
    link_latency_s: float,
) -> float:
    return (n_participants - 1) * link_latency_s


def _ring_ag_or_rs_bw_term_s(
    size: Size,
    n_participants: int,
    # NOTE: assumes duplex bw = 2x unidirectional
    unidirectional_link_bw_bytes_per_sec: int,
) -> float:
    phase_sent_data_size_bytes = int(size.bytes() / n_participants)
    return (n_participants - 1) * (
        phase_sent_data_size_bytes / unidirectional_link_bw_bytes_per_sec
    )


def _get_ring_tp_ag_or_rs_comm_time_s(
    size: Size,
    n_participants: int,
    machine_spec: MachineSpec,
    parallel_config: ParallelConfig,
    is_expert_comm: bool = False,
) -> float:
    lat_term_s = _ring_ag_or_rs_latency_term_s(
        n_participants=n_participants,
        # TODO. need to fix
        link_latency_s=machine_spec.intra_node_connect.latency_sec,
    )
    bw_term_s = _ring_ag_or_rs_bw_term_s(
        size,
        n_participants=n_participants,
        unidirectional_link_bw_bytes_per_sec=int(
            _get_effective_bw(
                parallelism_type=ParallelismType.ETP if is_expert_comm else ParallelismType.TP,
                parallel_config=parallel_config,
                machine_spec=machine_spec,
                is_expert_comm=is_expert_comm,
            )
        ),
    )

    return bw_term_s + lat_term_s


def get_all_to_all_comm_time_s(
    size: Size,
    parallel_config: ParallelConfig,
    machine_spec: MachineSpec,
) -> float:
    # For all-to-all in MoE context, participants are expert parallel ranks
    n_participants = parallel_config.expert_mesh.ep if parallel_config.expert_mesh else 1
    mp_degree_in_node = parallel_config.tp

    lat_term_s = machine_spec.inter_node_connect.latency_sec

    # we'll just model it as simultaneous sends of partition to all other participants.
    bw = (
        (machine_spec.inter_node_connect.unidirectional_bw_bytes_per_sec / mp_degree_in_node)
        if n_participants > 8
        else machine_spec.intra_node_connect.unidirectional_bw_bytes_per_sec
    )

    bw_term_s = ((size.bytes() // n_participants) * (n_participants - 1)) / bw

    return lat_term_s + bw_term_s


def get_expert_tp_all_gather_comm_time_s(
    size: Size,
    parallel_config: ParallelConfig,
    machine_spec: MachineSpec,
) -> float:
    """All-gather for expert tensor parallelism. Uses expert_tp degree."""
    if not parallel_config.expert_mesh:
        return 0.0
    return _get_ring_tp_ag_or_rs_comm_time_s(
        size,
        n_participants=parallel_config.expert_mesh.tp,
        machine_spec=machine_spec,
        parallel_config=parallel_config,
        is_expert_comm=True,
    )


def get_expert_tp_reduce_scatter_comm_time_s(
    size: Size,
    parallel_config: ParallelConfig,
    machine_spec: MachineSpec,
) -> float:
    """Reduce-scatter for expert tensor parallelism. Uses expert_tp degree."""
    if not parallel_config.expert_mesh:
        return 0.0
    return _get_ring_tp_ag_or_rs_comm_time_s(
        size,
        n_participants=parallel_config.expert_mesh.tp,
        machine_spec=machine_spec,
        parallel_config=parallel_config,
        is_expert_comm=True,
    )


def get_cross_dc_dp_all_gather_comm_time_s(
    size: Size,
    parallel_config: ParallelConfig,
    machine_spec: MachineSpec,
    cross_dc_config: CrossDCConfig,
) -> float:
    """Calculate cross-DC data parallel all-gather time.

    Args:
        size: Size of data to communicate
        parallel_config: Parallelism configuration
        machine_spec: Machine/hardware specification
        cross_dc_config: Cross-DC configuration

    Returns:
        Time in seconds for cross-DC all-gather
    """
    return _get_cross_dc_dp_all_gather_or_reduce_scatter_comm_time_s(
        size=size,
        parallel_config=parallel_config,
        machine_spec=machine_spec,
        cross_dc_config=cross_dc_config,
    )


def get_cross_dc_dp_reduce_scatter_comm_time_s(
    size: Size,
    parallel_config: ParallelConfig,
    machine_spec: MachineSpec,
    cross_dc_config: CrossDCConfig,
) -> float:
    """Calculate cross-DC data parallel reduce-scatter time.

    Args:
        size: Size of data to communicate
        parallel_config: Parallelism configuration
        machine_spec: Machine/hardware specification
        cross_dc_config: Cross-DC configuration

    Returns:
        Time in seconds for cross-DC reduce-scatter
    """
    return _get_cross_dc_dp_all_gather_or_reduce_scatter_comm_time_s(
        size=size,
        parallel_config=parallel_config,
        machine_spec=machine_spec,
        cross_dc_config=cross_dc_config,
    )


def _get_cross_dc_dp_all_gather_or_reduce_scatter_comm_time_s(
    size: Size,
    parallel_config: ParallelConfig,
    machine_spec: MachineSpec,
    cross_dc_config: CrossDCConfig,
) -> float:
    # we didn't forget about the MoE case.
    # we typically want to avoid doing expert DP across the DCs, as usually:
    # a) expert DP < non-expert DP, so doing it across DCs leads to more
    #    rings sharing the same link.
    # b) the majority of parameters are expert parameters, meaning more
    #    data to move across a limited link
    dp_degree = parallel_config.dp
    n_dp_rings = parallel_config.world_size() // dp_degree
    cross_dc_bw_per_ring = cross_dc_config.interconnect_bandwidth_bytes_per_sec() / n_dp_rings

    # cross-DC DP means creating heterogeneous rings, where the majority of links
    # are fast(er) inter-node links, and some of them are slower cross-DC links.
    # we use the lowest BW link when calculating how long ring collectives take.
    # this is because we expect the slowest link to behave like a dam, and
    # rate limit any faster links that come after it.
    effective_bw = min(
        cross_dc_bw_per_ring,
        machine_spec.inter_node_connect.unidirectional_bw_bytes_per_sec / machine_spec.n_devices,
    )
    bw_term = _ring_ag_or_rs_bw_term_s(
        size=size,
        n_participants=dp_degree,
        unidirectional_link_bw_bytes_per_sec=int(effective_bw),
    )

    # in a ring, there are (dp_degree - 1) total hops
    # n_dcs of these are inter-DC hops, the rest are intra-DC hops
    inter_dc_latency_term = cross_dc_config.n_dcs * cross_dc_config.interconnect_latency_s
    intra_dc_latency_term = (dp_degree - 1 - cross_dc_config.n_dcs) * (
        machine_spec.inter_node_connect.latency_sec
    )
    latency_term = inter_dc_latency_term + intra_dc_latency_term

    return latency_term + bw_term



"""
evaluates how optimal a given training job's physical topology is (in terms of
network hops in each DP ring).
"""

from argparse import ArgumentParser

import boto3  # type: ignore[import-untyped, import-not-found]

import dlcalc.utils.cluster.ec2
import dlcalc.utils.cluster.kubernetes
import dlcalc.utils.cluster.topology
from dlcalc.utils.cluster.topology import TreeNode
from dlcalc.utils.math import safe_divide


def main() -> None:
    parser = ArgumentParser(__doc__)
    parser.add_argument(
        "job_search_prefix",
        help="kubernetes pod name prefix to use when filtering pods returned by `kubectl get pods`",
    )
    parser.add_argument(
        "-t",
        "--tp-degree",
        type=int,
        required=True,
        help="tensor parallel (TP) degree",
    )
    parser.add_argument(
        "-p",
        "--pp-degree",
        type=int,
        required=True,
        help="pipeline parallel (PP) degree",
    )
    parser.add_argument(
        "-d",
        "--dp-degree",
        type=int,
        required=True,
        help="data parallel (DP) degree",
    )
    parser.add_argument(
        "-n",
        "--accelerators-per-node",
        type=int,
        default=8,
        help="number of accelerators per node",
    )
    args = parser.parse_args()
    job_search_prefix = args.job_search_prefix
    tp_degree = args.tp_degree
    pp_degree = args.pp_degree
    dp_degree = args.dp_degree
    accelerators_per_node = args.accelerators_per_node

    world_size = tp_degree * pp_degree * dp_degree
    expected_n_nodes = safe_divide(world_size, accelerators_per_node)

    assert tp_degree <= accelerators_per_node

    # we only care to optimize inter-node DP communication
    # (intra-node DP communication can't be optimized with better instance placement),
    # so we'll ignore the parts of the ring that traverse intra-node links.
    dp_ranks_per_node = safe_divide(accelerators_per_node, tp_degree)
    nodes_per_dp_ring = safe_divide(dp_degree, dp_ranks_per_node)
    print(
        f"DP={dp_degree} and {dp_ranks_per_node} DP ranks/node, "
        f"therefore there are {nodes_per_dp_ring} nodes per DP ring"
    )

    # everything we look at here is on the instance level.

    # if TP < accelerators per node, then there will be multiple DP ranks on a single
    # node. we'll ignore this intra-node DP communication as we just care about
    # improving inter-node DP communication
    # we're making the assumption that all TP groups are within a single node.
    # since we're looking at things at the instance level, there's only one
    # ring per PP rank.

    ######################################################################################
    # STEP 1: Get instance info involved in a kubernetes job
    ######################################################################################
    cluster_members = dlcalc.utils.cluster.kubernetes.get_kubernetes_cluster_members(
        job_search_prefix
    )
    if len(cluster_members) != expected_n_nodes:
        raise ValueError(
            f"with world size {world_size} "
            f"expected to find {expected_n_nodes} nodes "
            f"but found {len(cluster_members)}"
        )
    worker_id_to_instance_id = {m.worker_id: m.node_instance_id for m in cluster_members}
    (cluster_region,) = {m.node_region for m in cluster_members}
    cluster_instance_types = {m.node_instance_type for m in cluster_members}
    cluster_azs = {m.node_az for m in cluster_members}

    print(
        f"found {len(cluster_members)} matching pods, with "
        f"region: {cluster_region}, instance_types: {cluster_instance_types}, azs: {cluster_azs}"
    )

    node_id_to_node = dlcalc.utils.cluster.topology.build_instance_tree(
        ec2_client=boto3.client("ec2", region_name=cluster_region),
        accepted_node_instance_types=cluster_instance_types,
        accepted_node_availability_zones=cluster_azs,
        accepted_instance_ids=set(worker_id_to_instance_id.values()),
    )

    ######################################################################################
    # STEP 2: Determine connectivity path of DP rings
    ######################################################################################

    # find placement on physical topology
    for pp_rank in range(pp_degree):
        dp_ring_for_pp_rank: list[TreeNode] = []
        for dp_rank in range(nodes_per_dp_ring):
            worker_id = pp_rank * nodes_per_dp_ring + dp_rank
            instance_id = worker_id_to_instance_id[worker_id]
            dp_ring_for_pp_rank.append(node_id_to_node[instance_id])

        traversal_dist = dlcalc.utils.cluster.topology.traversal_distance_of_ring(
            dp_ring_for_pp_rank
        )
        print(
            f"DP ring for PP={pp_rank}: traversal distance: {traversal_dist} "
            f"-> avg traversal distance: {traversal_dist / nodes_per_dp_ring:.2f}"
        )
        print()


if __name__ == "__main__":
    main()



"""Unit tests for math.py utility functions.

This module tests mathematical utility functions used throughout
the dlcalc package for safe division, ceiling division, and products.

Test Coverage:
- safe_divide: Even division, error cases, edge cases
- ceil_divide: Ceiling division with various inputs
- product: Multiple arguments, edge cases
"""

import pytest

from dlcalc.utils.math import ceil_divide, product, safe_divide


class TestSafeDivide:
    """Test cases for the safe_divide function."""

    def test_even_division(self):
        """Test safe_divide with evenly divisible numbers."""
        assert safe_divide(10, 2) == 5
        assert safe_divide(100, 10) == 10
        assert safe_divide(1000, 50) == 20
        assert safe_divide(0, 5) == 0

    def test_division_by_one(self):
        """Test safe_divide with denominator of 1."""
        assert safe_divide(1, 1) == 1
        assert safe_divide(100, 1) == 100
        assert safe_divide(999, 1) == 999

    def test_same_numerator_denominator(self):
        """Test safe_divide when numerator equals denominator."""
        assert safe_divide(5, 5) == 1
        assert safe_divide(100, 100) == 1
        assert safe_divide(1234, 1234) == 1

    def test_uneven_division_raises_error(self):
        """Test that uneven division raises ValueError."""
        with pytest.raises(ValueError, match="10 not divisible by 3"):
            safe_divide(10, 3)

        with pytest.raises(ValueError, match="100 not divisible by 7"):
            safe_divide(100, 7)

    def test_error_message_includes_valid_denominators(self):
        """Test that error message includes list of valid denominators."""
        with pytest.raises(ValueError) as exc_info:
            safe_divide(12, 5)

        error_msg = str(exc_info.value)
        assert "12 not divisible by 5" in error_msg
        assert "valid denominators: [1, 2, 3, 4, 6]" in error_msg

    def test_large_numbers(self):
        """Test safe_divide with large numbers."""
        assert safe_divide(1000000, 1000) == 1000
        assert safe_divide(1024 * 1024, 1024) == 1024

        # Large number not divisible
        with pytest.raises(ValueError):
            safe_divide(1000001, 1000)

    def test_prime_numbers(self):
        """Test safe_divide with prime numbers."""
        # Prime number only divisible by 1 and itself
        assert safe_divide(17, 1) == 17
        assert safe_divide(17, 17) == 1

        with pytest.raises(ValueError) as exc_info:
            safe_divide(17, 5)

        # Prime numbers have only 1 as valid divisor (besides themselves)
        assert "valid denominators: [1]" in str(exc_info.value)

    def test_negative_numbers(self):
        """Test safe_divide behavior with negative numbers."""
        # Python's modulo works differently with negatives
        # This tests current behavior
        assert safe_divide(-10, 2) == -5
        assert safe_divide(10, -2) == -5
        assert safe_divide(-10, -2) == 5

    def test_powers_of_two(self):
        """Test safe_divide with powers of two (common in ML contexts)."""
        assert safe_divide(1024, 2) == 512
        assert safe_divide(1024, 4) == 256
        assert safe_divide(1024, 8) == 128
        assert safe_divide(1024, 16) == 64
        assert safe_divide(1024, 32) == 32


class TestCeilDivide:
    """Test cases for the ceil_divide function."""

    def test_even_division(self):
        """Test ceil_divide when result is exact."""
        assert ceil_divide(10, 2) == 5
        assert ceil_divide(100, 10) == 10
        assert ceil_divide(1000, 50) == 20

    def test_uneven_division_rounds_up(self):
        """Test ceil_divide rounds up for uneven division."""
        assert ceil_divide(10, 3) == 4  # 3.333... -> 4
        assert ceil_divide(7, 2) == 4  # 3.5 -> 4
        assert ceil_divide(1, 2) == 1  # 0.5 -> 1
        assert ceil_divide(99, 10) == 10  # 9.9 -> 10

    def test_division_by_one(self):
        """Test ceil_divide with denominator of 1."""
        assert ceil_divide(1, 1) == 1
        assert ceil_divide(100, 1) == 100
        assert ceil_divide(999, 1) == 999

    def test_zero_numerator(self):
        """Test ceil_divide with zero numerator."""
        assert ceil_divide(0, 1) == 0
        assert ceil_divide(0, 10) == 0
        assert ceil_divide(0, 100) == 0

    def test_large_numbers(self):
        """Test ceil_divide with large numbers."""
        assert ceil_divide(1000000, 999) == 1002  # ceiling of 1001.001...
        assert ceil_divide(1024 * 1024 + 1, 1024) == 1025

    def test_numerator_less_than_denominator(self):
        """Test ceil_divide when numerator < denominator."""
        assert ceil_divide(1, 10) == 1
        assert ceil_divide(5, 10) == 1
        assert ceil_divide(9, 10) == 1
        assert ceil_divide(10, 10) == 1
        assert ceil_divide(11, 10) == 2

    def test_negative_numbers(self):
        """Test ceil_divide with negative numbers."""
        # Python's math.ceil behavior with negatives
        assert ceil_divide(-10, 3) == -3  # -3.333... -> -3
        assert ceil_divide(10, -3) == -3  # -3.333... -> -3
        assert ceil_divide(-10, -3) == 4  # 3.333... -> 4

    def test_comparison_with_safe_divide(self):
        """Test that ceil_divide matches safe_divide for even divisions."""
        test_cases = [(100, 10), (1024, 32), (1000, 50)]
        for num, denom in test_cases:
            assert ceil_divide(num, denom) == safe_divide(num, denom)

    def test_floating_point_precision(self):
        """Test ceil_divide handles floating point correctly."""
        # These should all round up to the next integer
        assert ceil_divide(1000001, 1000000) == 2
        assert ceil_divide(100000000000001, 100000000000000) == 2


class TestProduct:
    """Test cases for the product function."""

    def test_single_argument(self):
        """Test product with single argument."""
        assert product(5) == 5
        assert product(0) == 0
        assert product(1) == 1
        assert product(-5) == -5

    def test_two_arguments(self):
        """Test product with two arguments."""
        assert product(2, 3) == 6
        assert product(10, 10) == 100
        assert product(0, 100) == 0

    def test_multiple_arguments(self):
        """Test product with multiple arguments."""
        assert product(2, 3, 4) == 24
        assert product(1, 2, 3, 4, 5) == 120
        assert product(10, 10, 10) == 1000

    def test_with_zero(self):
        """Test product when one argument is zero."""
        assert product(0, 1, 2, 3) == 0
        assert product(100, 0, 200) == 0
        assert product(1, 2, 3, 0, 4, 5) == 0

    def test_with_ones(self):
        """Test product with ones."""
        assert product(1, 1, 1, 1) == 1
        assert product(5, 1, 3, 1) == 15
        assert product(1, 10) == 10

    def test_negative_numbers(self):
        """Test product with negative numbers."""
        assert product(-2, 3) == -6
        assert product(-2, -3) == 6
        assert product(-1, -1, -1) == -1
        assert product(-2, -2, -2) == -8
        assert product(-2, 3, -4) == 24

    def test_large_numbers(self):
        """Test product with large numbers."""
        assert product(1000, 1000) == 1000000
        assert product(100, 100, 100) == 1000000

        # Common ML dimensions
        assert product(32, 1024, 768) == 25165824  # batch * seq * hidden

    def test_empty_arguments_assertion(self):
        """Test that product with no arguments raises assertion."""
        with pytest.raises(AssertionError):
            product()

    def test_powers_of_two(self):
        """Test product with powers of two (common in ML)."""
        assert product(2, 2, 2, 2) == 16
        assert product(4, 8, 16) == 512
        assert product(32, 64) == 2048

    def test_factorials_equivalent(self):
        """Test product can compute factorials."""
        # 5! = 5 * 4 * 3 * 2 * 1
        assert product(5, 4, 3, 2, 1) == 120
        # 6!
        assert product(6, 5, 4, 3, 2, 1) == 720

    def test_commutative_property(self):
        """Test that product is commutative."""
        assert product(2, 3, 4) == product(4, 3, 2)
        assert product(5, 10, 2) == product(2, 5, 10)

    def test_associative_property(self):
        """Test that product is associative."""
        # (2 * 3) * 4 = 2 * (3 * 4)
        assert product(product(2, 3), 4) == product(2, product(3, 4))

    def test_unpacking_lists(self):
        """Test product with unpacked lists."""
        numbers = [2, 3, 4, 5]
        assert product(*numbers) == 120

        dimensions = [32, 128, 768]  # batch, seq_len, hidden_dim
        assert product(*dimensions) == 3145728


class TestIntegration:
    """Integration tests for math utility functions."""

    def test_tensor_dimension_calculations(self):
        """Test realistic tensor dimension calculations."""
        # Calculate total elements in a partitioned tensor
        batch_size = 32
        seq_len = 1024
        hidden_dim = 768
        tp_degree = 8

        total_elements = product(batch_size, seq_len, hidden_dim)
        elements_per_partition = safe_divide(total_elements, tp_degree)

        assert total_elements == 25165824
        assert elements_per_partition == 3145728

    def test_memory_page_calculations(self):
        """Test memory page calculations with ceiling division."""
        page_size = 4096  # 4KB pages

        # Small allocation needs 1 page
        assert ceil_divide(100, page_size) == 1

        # Exact page boundary
        assert ceil_divide(4096, page_size) == 1
        assert ceil_divide(8192, page_size) == 2

        # Just over page boundary
        assert ceil_divide(4097, page_size) == 2
        assert ceil_divide(8193, page_size) == 3

    def test_batch_splitting(self):
        """Test batch splitting across devices."""
        global_batch_size = 512
        num_gpus = 8

        # Even split
        per_gpu_batch = safe_divide(global_batch_size, num_gpus)
        assert per_gpu_batch == 64

        # Uneven split should fail with safe_divide
        with pytest.raises(ValueError):
            safe_divide(513, num_gpus)

        # But works with ceil_divide
        assert ceil_divide(513, num_gpus) == 65

    def test_parameter_sharding(self):
        """Test parameter sharding calculations."""
        # Transformer layer dimensions
        vocab_size = 50000
        hidden_dim = 1024
        num_heads = 16

        # Total parameters in embedding
        embedding_params = product(vocab_size, hidden_dim)

        # Shard across tensor parallel dimension
        tp_degree = 4
        params_per_shard = safe_divide(embedding_params, tp_degree)

        assert embedding_params == 51200000
        assert params_per_shard == 12800000

        # Attention head splitting
        head_dim = safe_divide(hidden_dim, num_heads)
        assert head_dim == 64

    def test_combined_operations(self):
        """Test combining multiple math operations."""
        # Calculate memory for batch of sequences
        batch_size = 32
        seq_len = 2048
        hidden_dim = 768
        bytes_per_element = 2  # fp16

        # Total elements
        total_elements = product(batch_size, seq_len, hidden_dim)

        # Total bytes
        total_bytes = product(total_elements, bytes_per_element)

        # Divide into chunks of 1MB (1024 * 1024 bytes)
        chunk_size = product(1024, 1024)
        num_chunks = ceil_divide(total_bytes, chunk_size)

        assert total_elements == 50331648
        assert total_bytes == 100663296
        assert num_chunks == 96  # ~96MB



"""Integration tests for 3dtrn with cross-DC support."""

import re
import subprocess
import tempfile
from pathlib import Path

import pytest
import yaml


def run_3dtrn(config_file: str, timeout: int = 30) -> tuple[float | None, float | None, str]:
    """Run 3dtrn and extract MFU and total memory values from output.

    Args:
        config_file: Path to YAML configuration file
        timeout: Maximum time to wait for command completion

    Returns:
        Tuple of (MFU percentage or None, total memory in GiB or None, full output)
    """
    try:
        result = subprocess.run(
            ["3dtrn", config_file], capture_output=True, text=True, timeout=timeout
        )

        output = result.stdout + result.stderr

        # Find MFU in output - looking for "Theoretical MFU: XX.XX%" (with potential ANSI codes)
        # Remove ANSI escape codes first for easier matching
        clean_output = re.sub(r"\x1b\[[0-9;]*m", "", output)
        mfu_match = re.search(r"Theoretical MFU:\s*([0-9]+\.[0-9]+)%", clean_output)
        mfu_value = float(mfu_match.group(1)) if mfu_match else None

        # Find total memory in output - looking for "Total Memory Required: XX.XXX GiB"
        mem_match = re.search(r"Total Memory Required:\s*([0-9]+\.[0-9]+)\s*GiB", clean_output)
        mem_value = float(mem_match.group(1)) if mem_match else None

        return mfu_value, mem_value, output

    except subprocess.TimeoutExpired:
        pytest.fail(f"Timeout ({timeout}s) running {config_file}")
    except FileNotFoundError:
        pytest.skip("3dtrn command not found - install package first")
    except Exception as e:
        pytest.fail(f"Error running {config_file}: {e}")


class TestCrossDCIntegration:
    """Integration tests for cross-DC training performance modeling."""

    def test_cross_dc_config_parsing(self):
        """Test that 3dtrn correctly parses and uses cross_dc configuration."""
        # Create a test config with cross-DC parameters
        config = {
            "model": {
                "n_layers": 32,
                "hidden_sz": 4096,
                "inter_sz": 11008,
                "n_q_heads": 32,
                "n_kv_heads": 8,
                "head_dim": 128,
                "vocab_sz": 32000,
                "glu": True,
                "rotary_embeds": True,
                "dropout": False,
                "tie_embeddings": True,
            },
            "parallelism": {
                "tp": 8,
                "pp": 2,
                "dp": 16,
                "vpp": 1,
                "sp": True,
                "zero_level": 1,
                "n_param_buckets": 5,
            },
            "performance": {
                "activation_checkpointing_type": "selective",
            },
            "data": {
                "gbs": 256,
                "seqlen": 2048,
                "microbatch_sz": 1,
            },
            "hardware": {
                "node_type": "p5.48xlarge",
            },
            "cross_dc": {
                "n_dcs": 3,
                "interconnect_bandwidth_gbps": 800,
                "interconnect_latency_s": 0.0035,
            },
        }

        # Write config to temporary file
        with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as f:
            yaml.dump(config, f)
            config_path = f.name

        try:
            # Run 3dtrn with cross-DC config
            mfu, memory, output = run_3dtrn(config_path)

            # Check that output contains cross-DC information
            assert "cross-DC" in output or "Cross-DC" in output

            # MFU should be lower than baseline due to cross-DC overhead
            # (We can't check exact values without running baseline, but it should be reasonable)
            assert mfu is not None
            assert 0 < mfu < 100  # MFU should be between 0 and 100%

        finally:
            # Clean up temporary file
            Path(config_path).unlink()

    def test_cross_dc_degradation_reporting(self):
        """Test that cross-DC configuration shows degradation in output."""
        # Configuration with cross-DC
        config = {
            "model": {
                "n_layers": 32,
                "hidden_sz": 4096,
                "inter_sz": 11008,
                "n_q_heads": 32,
                "n_kv_heads": 8,
                "head_dim": 128,
                "vocab_sz": 32000,
                "glu": True,
                "rotary_embeds": True,
                "dropout": False,
                "tie_embeddings": True,
            },
            "parallelism": {
                "tp": 8,
                "pp": 2,
                "dp": 16,
                "vpp": 1,
                "sp": True,
                "zero_level": 1,
                "n_param_buckets": 5,
            },
            "performance": {
                "activation_checkpointing_type": "selective",
            },
            "data": {
                "gbs": 256,
                "seqlen": 2048,
                "microbatch_sz": 1,
            },
            "hardware": {
                "node_type": "p5.48xlarge",
            },
            "cross_dc": {
                "n_dcs": 3,
                "interconnect_bandwidth_gbps": 800,
                "interconnect_latency_s": 0.0035,
            },
        }

        # Run with cross-DC
        with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as f:
            yaml.dump(config, f)
            config_path = f.name

        try:
            mfu, _, output = run_3dtrn(config_path)
        finally:
            Path(config_path).unlink()

        # Verify cross-DC impact is reported
        assert mfu is not None
        assert "Cross-DC Impact" in output or "cross-DC" in output
        assert "degradation" in output.lower() or "slower" in output.lower()

        # Should show specific cross-DC metrics
        assert "Cross-DC Configuration" in output or "cross_dc" in output
        assert "800" in output  # bandwidth
        assert "3.5" in output or "0.0035" in output  # latency

    def test_cross_dc_output_format(self):
        """Test that cross-DC output includes expected information."""
        config = {
            "model": {
                "n_layers": 16,
                "hidden_sz": 2048,
                "inter_sz": 5504,
                "n_q_heads": 16,
                "n_kv_heads": 8,
                "head_dim": 128,
                "vocab_sz": 32000,
                "glu": True,
                "rotary_embeds": True,
                "dropout": False,
                "tie_embeddings": True,
            },
            "parallelism": {
                "tp": 4,
                "pp": 2,
                "dp": 8,
                "vpp": 1,
                "sp": True,
                "zero_level": 1,
                "n_param_buckets": 5,
            },
            "performance": {
                "activation_checkpointing_type": "selective",
            },
            "data": {
                "gbs": 64,
                "seqlen": 1024,
                "microbatch_sz": 1,
            },
            "hardware": {
                "node_type": "p4d.24xlarge",
            },
            "cross_dc": {
                "n_dcs": 2,
                "interconnect_bandwidth_gbps": 400,
                "interconnect_latency_s": 0.005,
            },
        }

        with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as f:
            yaml.dump(config, f)
            config_path = f.name

        try:
            mfu, memory, output = run_3dtrn(config_path)

            # Check for expected cross-DC output elements
            assert "Cross-DC Configuration" in output or "cross_dc" in output
            assert "degradation" in output.lower()

            # Should show bandwidth and latency information
            assert "800" in output or "400" in output  # bandwidth
            assert "0.005" in output or "5" in output  # latency

        finally:
            Path(config_path).unlink()



"""Unit tests for cross-DC configuration and communication functions."""

from dlcalc.utils.comms import (
    get_cross_dc_dp_all_gather_comm_time_s,
    get_cross_dc_dp_reduce_scatter_comm_time_s,
)
from dlcalc.utils.configurations import CrossDCConfig
from dlcalc.utils.data import Size
from dlcalc.utils.hardware import MachineSpec
from dlcalc.utils.model_3d import ParallelConfig


class TestCrossDCConfig:
    """Test CrossDCConfig dataclass."""

    def test_initialization(self):
        """Test CrossDCConfig can be initialized with required parameters."""
        config = CrossDCConfig(
            n_dcs=3,
            interconnect_bandwidth_gbps=800,
            interconnect_latency_s=0.0035,
        )
        assert config.n_dcs == 3
        assert config.interconnect_bandwidth_gbps == 800
        assert config.interconnect_latency_s == 0.0035


class TestCrossDCCommunication:
    """Test cross-DC communication functions."""

    def test_cross_dc_dp_reduce_scatter(self):
        """Test cross-DC data parallel reduce-scatter calculation."""
        # Setup test configuration
        cross_dc_config = CrossDCConfig(
            n_dcs=3,
            interconnect_bandwidth_gbps=800,
            interconnect_latency_s=0.0035,
        )

        parallel_config = ParallelConfig(
            tp=8,
            cp=1,
            pp=2,
            dp=64,
            expert_mesh=None,
            vpp=1,
            sp_enabled=True,
            zero_level=ParallelConfig.ZeroLevel.PARTITION_OPTIMIZER,
        )

        machine_spec = MachineSpec.from_str("p5.48xlarge")

        # Size of data to communicate (e.g., gradient bucket)
        size = Size(numel=250_000_000, bits_per_element=32)  # 250M params, 32-bit grads

        # Calculate cross-DC reduce-scatter time
        cross_dc_time = get_cross_dc_dp_reduce_scatter_comm_time_s(
            size=size,
            parallel_config=parallel_config,
            machine_spec=machine_spec,
            cross_dc_config=cross_dc_config,
        )

        # Should be significantly higher than baseline due to cross-DC bottleneck
        assert cross_dc_time > 0
        # Rough check that it's in a reasonable range (milliseconds to seconds)
        assert 0.001 < cross_dc_time < 10.0

    def test_cross_dc_dp_all_gather(self):
        """Test cross-DC data parallel all-gather calculation."""
        cross_dc_config = CrossDCConfig(
            n_dcs=3,
            interconnect_bandwidth_gbps=800,
            interconnect_latency_s=0.0035,
        )

        parallel_config = ParallelConfig(
            tp=8,
            cp=1,
            pp=2,
            dp=64,
            expert_mesh=None,
            vpp=1,
            sp_enabled=True,
            zero_level=ParallelConfig.ZeroLevel.PARTITION_OPTIMIZER,
        )

        machine_spec = MachineSpec.from_str("p5.48xlarge")

        # Size of data to communicate (e.g., parameter bucket)
        size = Size(numel=250_000_000, bits_per_element=16)  # 250M params, 16-bit

        # Calculate cross-DC all-gather time
        cross_dc_time = get_cross_dc_dp_all_gather_comm_time_s(
            size=size,
            parallel_config=parallel_config,
            machine_spec=machine_spec,
            cross_dc_config=cross_dc_config,
        )

        # Should be significantly higher than baseline due to cross-DC bottleneck
        assert cross_dc_time > 0
        # Rough check that it's in a reasonable range
        assert 0.001 < cross_dc_time < 10.0

    def test_cross_dc_vs_baseline_degradation(self):
        """Test that cross-DC communication is slower than baseline."""
        from dlcalc.utils.comms import (
            get_dp_all_gather_comm_time_s,
            get_dp_reduce_scatter_comm_time_s,
        )

        cross_dc_config = CrossDCConfig(
            n_dcs=3,
            interconnect_bandwidth_gbps=800,  # Limited cross-DC bandwidth
            interconnect_latency_s=0.0035,  # 3.5ms latency
        )

        parallel_config = ParallelConfig(
            tp=8,
            cp=1,
            pp=2,
            dp=64,
            expert_mesh=None,
            vpp=1,
            sp_enabled=True,
            zero_level=ParallelConfig.ZeroLevel.PARTITION_OPTIMIZER,
        )

        machine_spec = MachineSpec.from_str("p5.48xlarge")
        size = Size(numel=250_000_000, bits_per_element=32)

        # Calculate baseline (single DC) times
        baseline_rs = get_dp_reduce_scatter_comm_time_s(
            size=size,
            parallel_config=parallel_config,
            machine_spec=machine_spec,
        )

        baseline_ag = get_dp_all_gather_comm_time_s(
            size=Size(numel=250_000_000, bits_per_element=16),
            parallel_config=parallel_config,
            machine_spec=machine_spec,
        )

        # Calculate cross-DC times
        cross_dc_rs = get_cross_dc_dp_reduce_scatter_comm_time_s(
            size=size,
            parallel_config=parallel_config,
            machine_spec=machine_spec,
            cross_dc_config=cross_dc_config,
        )

        cross_dc_ag = get_cross_dc_dp_all_gather_comm_time_s(
            size=Size(numel=250_000_000, bits_per_element=16),
            parallel_config=parallel_config,
            machine_spec=machine_spec,
            cross_dc_config=cross_dc_config,
        )

        # Cross-DC should be slower than baseline
        assert cross_dc_rs > baseline_rs
        assert cross_dc_ag > baseline_ag

        # Degradation should be significant but not unreasonable
        # Cross-DC can be quite a bit slower due to bandwidth and latency constraints
        assert 1.1 < (cross_dc_rs / baseline_rs) < 100
        assert 1.1 < (cross_dc_ag / baseline_ag) < 100



"""Unit tests for data.py classes.

This module tests the Size and TensorRepr classes which are fundamental
data structures for representing memory sizes and distributed tensor shapes
in the dlcalc package.

Test Coverage:
- Size class: Basic operations, arithmetic, edge cases
- TensorRepr class: Partitioning, shape calculations, size methods
- Integration: Combined usage of Size and TensorRepr for realistic scenarios
"""

import pytest

from dlcalc.utils.data import Size, TensorRepr


class TestSize:
    """Test cases for the Size class."""

    def test_basic_initialization(self):
        """Test basic Size initialization."""
        size = Size(numel=1000, bits_per_element=16)
        assert size.numel() == 1000
        assert size._bits_per_element == 16

    def test_bits_calculation(self):
        """Test bits calculation."""
        size = Size(numel=1000, bits_per_element=16)
        assert size.bits() == 16000

    def test_bytes_calculation(self):
        """Test bytes calculation."""
        # 16 bits = 2 bytes per element
        size = Size(numel=1000, bits_per_element=16)
        assert size.bytes() == 2000

        # 32 bits = 4 bytes per element
        size = Size(numel=500, bits_per_element=32)
        assert size.bytes() == 2000

        # 8 bits = 1 byte per element
        size = Size(numel=100, bits_per_element=8)
        assert size.bytes() == 100

    def test_addition_same_bits(self):
        """Test addition of Size objects with same bits per element."""
        size1 = Size(numel=1000, bits_per_element=16)
        size2 = Size(numel=500, bits_per_element=16)
        result = size1 + size2

        assert result.numel() == 1500
        assert result._bits_per_element == 16
        assert result.bytes() == 3000

    def test_addition_different_bits_raises_error(self):
        """Test that adding sizes with different bits per element raises error."""
        size1 = Size(numel=1000, bits_per_element=16)
        size2 = Size(numel=500, bits_per_element=32)

        with pytest.raises(ValueError, match="different bits per element"):
            _ = size1 + size2

    def test_multiplication(self):
        """Test multiplication of Size by integer."""
        size = Size(numel=100, bits_per_element=16)

        # Left multiplication
        result = size * 3
        assert result.numel() == 300
        assert result._bits_per_element == 16

        # Right multiplication (rmul)
        result = 5 * size
        assert result.numel() == 500
        assert result._bits_per_element == 16

    def test_floor_division(self):
        """Test floor division of Size by integer."""
        size = Size(numel=1000, bits_per_element=16)
        result = size // 3

        assert result.numel() == 333  # floor division
        assert result._bits_per_element == 16

    def test_repr(self):
        """Test string representation of Size."""
        size = Size(numel=int(1e9), bits_per_element=16)
        repr_str = repr(size)

        assert "1.000 B" in repr_str  # 1B elements
        assert "1.863 GiB" in repr_str  # 2GB in GiB

    def test_edge_cases(self):
        """Test edge cases for Size."""
        # Zero elements
        size = Size(numel=0, bits_per_element=16)
        assert size.bits() == 0
        assert size.bytes() == 0

        # Large numbers
        size = Size(numel=int(1e12), bits_per_element=32)
        assert size.bits() == int(32e12)
        assert size.bytes() == int(4e12)

        # Chain operations
        size = Size(numel=100, bits_per_element=16)
        result = size * 2 + size * 3
        assert result.numel() == 500

        # Division by 1
        result = size // 1
        assert result.numel() == size.numel()


class TestTensorRepr:
    """Test cases for the TensorRepr class."""

    def test_basic_initialization(self):
        """Test basic TensorRepr initialization."""
        tensor = TensorRepr(unpartitioned_shape=(1024, 768), partition_spec={}, bits_per_elt=16)

        assert tensor._unpartitioned_shape == (1024, 768)
        assert tensor._partition_spec == {}
        assert tensor._bits_per_elt == 16

    def test_unpartitioned_shape_and_numel(self):
        """Test shape and numel for unpartitioned tensor."""
        tensor = TensorRepr(unpartitioned_shape=(100, 200, 300), partition_spec={}, bits_per_elt=32)

        assert tensor.shape(partitioned=False) == (100, 200, 300)
        assert tensor.numel(partitioned=False) == 100 * 200 * 300

    def test_partitioned_shape_single_axis(self):
        """Test shape for tensor partitioned along single axis."""
        tensor = TensorRepr(
            unpartitioned_shape=(1024, 768),
            partition_spec={0: 4},  # Partition first dimension by 4
            bits_per_elt=16,
        )

        assert tensor.shape(partitioned=False) == (1024, 768)
        assert tensor.shape(partitioned=True) == (256, 768)

    def test_partitioned_shape_multiple_axes(self):
        """Test shape for tensor partitioned along multiple axes."""
        tensor = TensorRepr(
            unpartitioned_shape=(1024, 768, 512),
            partition_spec={0: 4, 2: 2},  # Partition dim 0 by 4, dim 2 by 2
            bits_per_elt=16,
        )

        assert tensor.shape(partitioned=False) == (1024, 768, 512)
        assert tensor.shape(partitioned=True) == (256, 768, 256)

    def test_partitioned_numel(self):
        """Test numel calculation for partitioned tensor."""
        tensor = TensorRepr(
            unpartitioned_shape=(1000, 2000),
            partition_spec={0: 10, 1: 4},  # Total partitioning degree = 40
            bits_per_elt=32,
        )

        unpartitioned_numel = 1000 * 2000
        assert tensor.numel(partitioned=False) == unpartitioned_numel
        assert tensor.numel(partitioned=True) == unpartitioned_numel // 40

    def test_size_method(self):
        """Test the size method returns correct Size object."""
        tensor = TensorRepr(unpartitioned_shape=(100, 200), partition_spec={1: 2}, bits_per_elt=16)

        # Unpartitioned size
        size = tensor.size(partitioned=False)
        assert isinstance(size, Size)
        assert size.numel() == 20000
        assert size._bits_per_element == 16

        # Partitioned size
        size = tensor.size(partitioned=True)
        assert isinstance(size, Size)
        assert size.numel() == 10000
        assert size._bits_per_element == 16

    def test_enforce_evenly_partitionable_true(self):
        """Test that unevenly partitionable tensors raise error when enforced."""
        with pytest.raises(RuntimeError, match="not divisible"):
            TensorRepr(
                unpartitioned_shape=(100, 200),
                partition_spec={0: 3},  # 100 not divisible by 3
                bits_per_elt=16,
                enforce_evenly_partitionable=True,
            )

    def test_enforce_evenly_partitionable_false(self):
        """Test that unevenly partitionable tensors can be created when not enforced."""
        # Should not raise error during initialization
        tensor = TensorRepr(
            unpartitioned_shape=(100, 200),
            partition_spec={0: 3},  # 100 not divisible by 3
            bits_per_elt=16,
            enforce_evenly_partitionable=False,
        )

        # Note: safe_divide will still raise an error when actually calculating shape
        # This is the current behavior - the flag only affects initialization
        with pytest.raises(ValueError, match="not divisible"):
            _ = tensor.shape(partitioned=True)

    def test_empty_partition_spec(self):
        """Test tensor with empty partition spec behaves as unpartitioned."""
        tensor = TensorRepr(unpartitioned_shape=(512, 1024), partition_spec={}, bits_per_elt=32)

        assert tensor.shape(partitioned=True) == tensor.shape(partitioned=False)
        assert tensor.numel(partitioned=True) == tensor.numel(partitioned=False)

    def test_complex_partitioning(self):
        """Test complex partitioning scenario."""
        # 4D tensor partitioned along 3 dimensions
        tensor = TensorRepr(
            unpartitioned_shape=(64, 128, 256, 512),
            partition_spec={0: 2, 1: 4, 3: 8},  # Total degree = 64
            bits_per_elt=16,
        )

        assert tensor.shape(partitioned=True) == (32, 32, 256, 64)

        unpartitioned_numel = 64 * 128 * 256 * 512
        partitioned_numel = 32 * 32 * 256 * 64

        assert tensor.numel(partitioned=False) == unpartitioned_numel
        assert tensor.numel(partitioned=True) == partitioned_numel

    def test_single_element_tensor(self):
        """Test edge case of single element tensor."""
        tensor = TensorRepr(unpartitioned_shape=(1,), partition_spec={}, bits_per_elt=32)

        assert tensor.shape(partitioned=False) == (1,)
        assert tensor.numel(partitioned=False) == 1
        assert tensor.size(partitioned=False).bytes() == 4

    def test_high_dimensional_tensor(self):
        """Test tensor with many dimensions."""
        shape = (2, 3, 4, 5, 6, 7, 8, 9)
        tensor = TensorRepr(
            unpartitioned_shape=shape, partition_spec={0: 2, 3: 5, 7: 3}, bits_per_elt=8
        )

        expected_partitioned = (1, 3, 4, 1, 6, 7, 8, 3)
        assert tensor.shape(partitioned=True) == expected_partitioned

    def test_partition_spec_with_large_degrees(self):
        """Test partitioning with large partition degrees."""
        tensor = TensorRepr(
            unpartitioned_shape=(10000, 20000), partition_spec={0: 100, 1: 200}, bits_per_elt=16
        )

        assert tensor.shape(partitioned=True) == (100, 100)
        assert tensor.numel(partitioned=True) == 10000

    def test_bits_per_element_variations(self):
        """Test different bits per element values."""
        shape = (100, 100)

        # 8-bit tensor
        tensor8 = TensorRepr(shape, {}, bits_per_elt=8)
        assert tensor8.size(partitioned=False).bytes() == 10000

        # 16-bit tensor
        tensor16 = TensorRepr(shape, {}, bits_per_elt=16)
        assert tensor16.size(partitioned=False).bytes() == 20000

        # 32-bit tensor
        tensor32 = TensorRepr(shape, {}, bits_per_elt=32)
        assert tensor32.size(partitioned=False).bytes() == 40000

        # 64-bit tensor
        tensor64 = TensorRepr(shape, {}, bits_per_elt=64)
        assert tensor64.size(partitioned=False).bytes() == 80000


class TestIntegration:
    """Integration tests for Size and TensorRepr working together."""

    def test_tensor_size_operations(self):
        """Test operations on sizes from TensorRepr."""
        tensor1 = TensorRepr(unpartitioned_shape=(100, 100), partition_spec={}, bits_per_elt=16)
        tensor2 = TensorRepr(unpartitioned_shape=(50, 50), partition_spec={}, bits_per_elt=16)

        size1 = tensor1.size(partitioned=False)
        size2 = tensor2.size(partitioned=False)

        # Add sizes
        total_size = size1 + size2
        assert total_size.numel() == 12500  # 10000 + 2500
        assert total_size.bytes() == 25000  # 16 bits = 2 bytes per element

    def test_partitioned_tensor_memory_calculation(self):
        """Test realistic memory calculation for partitioned tensor."""
        # Simulating a large weight matrix partitioned for tensor parallelism
        tensor = TensorRepr(
            unpartitioned_shape=(4096, 8192),  # 32M parameters
            partition_spec={1: 8},  # TP=8
            bits_per_elt=16,  # fp16
        )

        # Unpartitioned memory
        full_size = tensor.size(partitioned=False)
        assert full_size.numel() == 4096 * 8192
        assert full_size.bytes() == 4096 * 8192 * 2  # 64MB

        # Partitioned memory (per device)
        shard_size = tensor.size(partitioned=True)
        assert shard_size.numel() == 4096 * 1024  # 4M parameters per device
        assert shard_size.bytes() == 4096 * 1024 * 2  # 8MB per device

    def test_multiple_tensor_total_memory(self):
        """Test calculating total memory for multiple tensors."""
        # Simulate transformer layer weights
        qkv_weight = TensorRepr(
            unpartitioned_shape=(768, 2304),  # 768 * 3 for Q, K, V
            partition_spec={1: 4},
            bits_per_elt=16,
        )

        mlp_weight = TensorRepr(
            unpartitioned_shape=(768, 3072), partition_spec={1: 4}, bits_per_elt=16
        )

        # Calculate total partitioned memory
        qkv_size = qkv_weight.size(partitioned=True)
        mlp_size = mlp_weight.size(partitioned=True)
        total_size = qkv_size + mlp_size

        expected_qkv_numel = 768 * 576  # 2304/4
        expected_mlp_numel = 768 * 768  # 3072/4

        assert qkv_size.numel() == expected_qkv_numel
        assert mlp_size.numel() == expected_mlp_numel
        assert total_size.numel() == expected_qkv_numel + expected_mlp_numel



"""Integration tests for 3dtrn MFU predictions.

This module tests that the 3dtrn calculator produces expected MFU values
for known configurations.
"""

import re
import subprocess
from pathlib import Path

import pytest


def run_3dtrn(config_file: str, timeout: int = 30) -> tuple[float | None, float | None, str]:
    """Run 3dtrn and extract MFU and total memory values from output.

    Args:
        config_file: Path to YAML configuration file
        timeout: Maximum time to wait for command completion

    Returns:
        Tuple of (MFU percentage or None, total memory in GiB or None, full output)
    """
    try:
        result = subprocess.run(
            ["3dtrn", config_file], capture_output=True, text=True, timeout=timeout
        )

        output = result.stdout + result.stderr

        # Find MFU in output - looking for "Theoretical MFU: XX.XX%" (with potential ANSI codes)
        # Remove ANSI escape codes first for easier matching
        clean_output = re.sub(r"\x1b\[[0-9;]*m", "", output)
        mfu_match = re.search(r"Theoretical MFU:\s*([0-9]+\.[0-9]+)%", clean_output)
        mfu_value = float(mfu_match.group(1)) if mfu_match else None

        # Find total memory in output - looking for "Total Memory Required: XX.XXX GiB"
        mem_match = re.search(r"Total Memory Required:\s*([0-9]+\.[0-9]+)\s*GiB", clean_output)
        mem_value = float(mem_match.group(1)) if mem_match else None

        return mfu_value, mem_value, output

    except subprocess.TimeoutExpired:
        pytest.fail(f"Timeout ({timeout}s) running {config_file}")
    except FileNotFoundError:
        pytest.skip("3dtrn command not found - install package first")
    except Exception as e:
        pytest.fail(f"Error running {config_file}: {e}")


class TestMFUPredictions:
    """Test MFU predictions for various model configurations."""

    def test_llama3_70b_mfu(self):
        """Test that LLaMA-3 70B configuration predicts ~30.22% MFU and correct memory."""
        config_file = "examples/llama3_70b.yaml"

        # Skip if config file doesn't exist
        if not Path(config_file).exists():
            pytest.skip(f"Config file {config_file} not found")

        # Run 3dtrn and get MFU and memory
        actual_mfu, actual_mem, output = run_3dtrn(config_file)

        # Check that we got an MFU value
        assert actual_mfu is not None, f"Could not find MFU in output:\n{output[-1000:]}"

        # Expected MFU value with tolerance
        expected_mfu = 30.22
        tolerance = 0.5  # Allow 0.5% difference

        # Check if MFU within tolerance
        difference = abs(actual_mfu - expected_mfu)
        assert difference <= tolerance, (
            f"MFU prediction {actual_mfu}% is outside tolerance. "
            f"Expected: {expected_mfu}% ± {tolerance}%, "
            f"Difference: {difference:.2f}%"
        )

        # Check that we got a memory value
        assert actual_mem is not None, f"Could not find total memory in output:\n{output[-1000:]}"

        # Expected memory value with tolerance
        expected_mem = 31.965  # GiB
        mem_tolerance = 0.5  # Allow 0.5 GiB difference

        # Check if memory within tolerance
        mem_difference = abs(actual_mem - expected_mem)
        assert mem_difference <= mem_tolerance, (
            f"Memory prediction {actual_mem} GiB is outside tolerance. "
            f"Expected: {expected_mem} GiB ± {mem_tolerance} GiB, "
            f"Difference: {mem_difference:.3f} GiB"
        )

    def test_gpt_oss_120b_mfu(self):
        """Test that GPT OSS 120B MoE configuration predicts the right MFU and memory."""
        config_file = "examples/gpt_oss_120b.yaml"

        # Skip if config file doesn't exist
        if not Path(config_file).exists():
            pytest.skip(f"Config file {config_file} not found")

        # Run 3dtrn and get MFU and memory
        actual_mfu, actual_mem, output = run_3dtrn(config_file)

        # Check that we got an MFU value
        assert actual_mfu is not None, f"Could not find MFU in output:\n{output[-1000:]}"

        # Expected MFU value with tolerance
        # Updated for alltoall time calculation fixes
        expected_mfu = 13.37
        tolerance = 0.5  # Allow 0.5% difference

        # Check if MFU within tolerance
        difference = abs(actual_mfu - expected_mfu)
        assert difference <= tolerance, (
            f"MFU prediction {actual_mfu}% is outside tolerance. "
            f"Expected: {expected_mfu}% ± {tolerance}%, "
            f"Difference: {difference:.2f}%"
        )

        # Check that we got a memory value
        assert actual_mem is not None, f"Could not find total memory in output:\n{output[-1000:]}"

        # Expected memory value with tolerance
        # Updated for model_3d changes
        expected_mem = 15.119  # GiB
        mem_tolerance = 0.5  # Allow 0.5 GiB difference

        # Check if memory within tolerance
        mem_difference = abs(actual_mem - expected_mem)
        assert mem_difference <= mem_tolerance, (
            f"Memory prediction {actual_mem} GiB is outside tolerance. "
            f"Expected: {expected_mem} GiB ± {mem_tolerance} GiB, "
            f"Difference: {mem_difference:.3f} GiB"
        )

        # Check that output mentions MoE since this is an MoE model
        assert "MoE" in output or "expert" in output.lower(), (
            "Output should mention MoE configuration"
        )

    def test_multiple_configs(self):
        """Test MFU predictions for multiple configurations if available."""
        # Define test cases: (config_file, expected_mfu, tolerance)
        test_cases = [
            ("examples/llama3_70b.yaml", 30.22, 0.5),
            ("examples/gpt_oss_120b.yaml", 13.37, 0.5),  # Updated for alltoall fixes
            # Add more test cases as needed
            # ("examples/llama3_8b.yaml", 45.0, 1.0),
        ]

        results = []
        for config_file, expected_mfu, tolerance in test_cases:
            if not Path(config_file).exists():
                results.append((config_file, "SKIPPED", "File not found"))
                continue

            actual_mfu, actual_mem, output = run_3dtrn(config_file)

            if actual_mfu is None:
                results.append((config_file, "FAILED", "MFU not found in output"))
                continue

            difference = abs(actual_mfu - expected_mfu)
            if difference <= tolerance:
                results.append((config_file, "PASSED", f"{actual_mfu}%"))
            else:
                results.append(
                    (
                        config_file,
                        "FAILED",
                        f"{actual_mfu}% (expected {expected_mfu}% ± {tolerance}%)",
                    )
                )

        # Print summary
        print("\nMFU Prediction Test Results:")
        print("-" * 60)
        for config, status, details in results:
            print(f"{config:40} {status:8} {details}")

        # Check if any failed
        failed = [r for r in results if r[1] == "FAILED"]
        if failed:
            pytest.fail(f"{len(failed)} configuration(s) failed MFU prediction test")

    def test_output_format(self):
        """Test that 3dtrn output contains expected sections."""
        config_file = "examples/llama3_70b.yaml"

        if not Path(config_file).exists():
            pytest.skip(f"Config file {config_file} not found")

        _, _, output = run_3dtrn(config_file)

        # Check for expected sections in output (remove ANSI codes first)
        clean_output = re.sub(r"\x1b\[[0-9;]*m", "", output)
        expected_sections = [
            "CONFIGURATION",
            "Model Architecture",
            "MEMORY",
            "Theoretical MFU:",
        ]

        missing_sections = []
        for section in expected_sections:
            if section not in clean_output:
                missing_sections.append(section)

        assert not missing_sections, (
            f"Output missing expected sections: {missing_sections}\nOutput sample:\n{output[:500]}"
        )

    @pytest.mark.parametrize(
        "batch_size,expected_range",
        [
            (2048, (25, 35)),  # Normal batch size
            (1024, (20, 30)),  # Smaller batch
            (4096, (30, 40)),  # Larger batch
        ],
    )
    def test_batch_size_impact(self, batch_size, expected_range, tmp_path):
        """Test that batch size affects MFU as expected."""
        # Create a temporary config with different batch sizes
        base_config = Path("examples/llama3_70b.yaml")

        if not base_config.exists():
            pytest.skip("Base config file not found")

        import yaml

        # Load base config
        with open(base_config) as f:
            config = yaml.safe_load(f)

        # Modify batch size
        config["data"]["gbs"] = batch_size

        # Save to temp file
        temp_config = tmp_path / f"test_bs_{batch_size}.yaml"
        with open(temp_config, "w") as f:
            yaml.dump(config, f)

        # Run and check MFU
        actual_mfu, actual_mem, output = run_3dtrn(str(temp_config))

        if actual_mfu is None:
            pytest.fail(f"Could not extract MFU for batch size {batch_size}")

        # Check if in expected range
        min_mfu, max_mfu = expected_range
        assert min_mfu <= actual_mfu <= max_mfu, (
            f"MFU {actual_mfu}% for batch size {batch_size} "
            f"outside expected range [{min_mfu}, {max_mfu}]"
        )


class TestErrorHandling:
    """Test error handling in 3dtrn."""

    def test_missing_file(self):
        """Test that 3dtrn handles missing files gracefully."""
        result = subprocess.run(["3dtrn", "non_existent_file.yaml"], capture_output=True, text=True)

        assert result.returncode != 0, "Should return non-zero exit code for missing file"

        error_output = result.stdout + result.stderr
        assert (
            "No such file" in error_output
            or "not found" in error_output
            or "does not exist" in error_output
        ), f"Should report file not found error, got: {error_output}"

    def test_invalid_yaml(self, tmp_path):
        """Test that 3dtrn handles invalid YAML gracefully."""
        # Create invalid YAML file
        invalid_yaml = tmp_path / "invalid.yaml"
        invalid_yaml.write_text("{ this is: invalid yaml ][")

        result = subprocess.run(["3dtrn", str(invalid_yaml)], capture_output=True, text=True)

        assert result.returncode != 0, "Should return non-zero exit code for invalid YAML"

        error_output = result.stdout + result.stderr
        assert "yaml" in error_output.lower() or "parse" in error_output.lower(), (
            f"Should report YAML parsing error, got: {error_output}"
        )

    def test_missing_required_fields(self, tmp_path):
        """Test that 3dtrn handles configs with missing required fields."""
        # Create YAML with missing fields
        incomplete_config = tmp_path / "incomplete.yaml"
        incomplete_config.write_text("""
model:
  n_layers: 32
# Missing other required fields
""")

        result = subprocess.run(["3dtrn", str(incomplete_config)], capture_output=True, text=True)

        assert result.returncode != 0, "Should return non-zero exit code for incomplete config"


@pytest.mark.slow
class TestPerformance:
    """Performance tests for 3dtrn (marked as slow)."""

    def test_execution_time(self):
        """Test that 3dtrn completes within reasonable time."""
        import time

        config_file = "examples/llama3_70b.yaml"

        if not Path(config_file).exists():
            pytest.skip(f"Config file {config_file} not found")

        start_time = time.time()
        run_3dtrn(config_file, timeout=10)
        execution_time = time.time() - start_time

        # Should complete within 5 seconds for a single config
        assert execution_time < 5, f"Execution took {execution_time:.2f}s, expected < 5s"


# Fixtures for common test data
@pytest.fixture
def sample_config(tmp_path):
    """Create a sample configuration for testing."""
    config = {
        "model": {
            "n_layers": 32,
            "hidden_sz": 4096,
            "inter_sz": 11008,
            "n_q_heads": 32,
            "n_kv_heads": 32,
            "head_dim": 128,
            "vocab_sz": 32000,
            "glu": True,
            "rotary_embeds": True,
            "dropout": False,
            "tie_embeddings": True,
        },
        "parallelism": {
            "tp": 4,
            "pp": 2,
            "dp": 16,
            "vpp": 1,
            "sp": True,
            "zero_level": 1,
            "n_param_buckets": 5,
        },
        "performance": {"activation_checkpointing_type": "selective"},
        "data": {"gbs": 512, "seqlen": 2048, "microbatch_sz": 1},
        "hardware": {"node_type": "p4d.24xlarge"},
    }

    config_file = tmp_path / "sample_config.yaml"
    import yaml

    with open(config_file, "w") as f:
        yaml.dump(config, f)

    return config_file


def test_with_sample_config(sample_config):
    """Test 3dtrn with a generated sample configuration."""
    actual_mfu, actual_mem, output = run_3dtrn(str(sample_config))

    # Should produce some MFU value
    assert actual_mfu is not None, "Should calculate MFU for sample config"
    assert 0 < actual_mfu < 100, f"MFU should be between 0 and 100, got {actual_mfu}"

    # Should produce some memory value
    assert actual_mem is not None, "Should calculate total memory for sample config"
    assert actual_mem > 0, f"Memory should be positive, got {actual_mem} GiB"

    # Should have standard output sections (check after removing ANSI codes)
    clean_output = re.sub(r"\x1b\[[0-9;]*m", "", output)
    assert "CONFIGURATION" in clean_output
    assert "Model Architecture" in clean_output
    assert "MEMORY" in clean_output



"""Unit tests for model_3d module, focusing on ParallelConfig."""

import pytest

from dlcalc.utils.model_3d import ParallelConfig


class TestParallelConfig:
    """Test ParallelConfig class functionality."""

    def test_basic_initialization(self):
        """Test basic ParallelConfig initialization without expert mesh."""
        config = ParallelConfig(
            tp=4,
            cp=2,
            pp=2,
            dp=8,
            expert_mesh=None,
            vpp=1,
            sp_enabled=True,
            zero_level=ParallelConfig.ZeroLevel.PARTITION_OPTIMIZER,
        )

        assert config.tp == 4
        assert config.cp == 2
        assert config.pp == 2
        assert config.dp == 8
        assert config.expert_mesh is None
        assert config.vpp == 1
        assert config.sp_enabled is True
        assert config.zero_level == ParallelConfig.ZeroLevel.PARTITION_OPTIMIZER

    def test_world_size_calculation(self):
        """Test world_size calculation."""
        config = ParallelConfig(
            tp=4,
            cp=2,
            pp=2,
            dp=8,
            expert_mesh=None,
            vpp=1,
            sp_enabled=False,
            zero_level=ParallelConfig.ZeroLevel.PARTITION_OPTIMIZER,
        )

        assert config.world_size() == 4 * 2 * 2 * 8  # tp * cp * pp * dp = 128

    def test_expert_parallel_cfg_initialization(self):
        """Test ExpertParallelCfg initialization."""
        expert_cfg = ParallelConfig.ExpertParallelCfg(ep=8, tp=2, dp=4)

        assert expert_cfg.ep == 8
        assert expert_cfg.tp == 2
        assert expert_cfg.dp == 4

    def test_parallel_config_with_valid_expert_mesh(self):
        """Test ParallelConfig with a valid expert mesh configuration."""
        # The constraint is: expert_mesh.ep * expert_mesh.tp * expert_mesh.dp == dp * cp * tp
        # With tp=4, cp=2, dp=8: dp * cp * tp = 8 * 2 * 4 = 64
        # So expert_mesh.ep * expert_mesh.tp * expert_mesh.dp must equal 64
        expert_mesh = ParallelConfig.ExpertParallelCfg(
            ep=8,  # 8 * 2 * 4 = 64 ✓
            tp=2,
            dp=4,
        )

        config = ParallelConfig(
            tp=4,
            cp=2,
            pp=2,
            dp=8,
            expert_mesh=expert_mesh,
            vpp=1,
            sp_enabled=True,
            zero_level=ParallelConfig.ZeroLevel.PARTITION_OPTIMIZER,
        )

        assert config.expert_mesh.ep == 8
        assert config.expert_mesh.tp == 2
        assert config.expert_mesh.dp == 4

    def test_parallel_config_with_invalid_expert_mesh_raises_error(self):
        """Test that invalid expert mesh configuration raises assertion error."""
        # The constraint is: expert_mesh.ep * expert_mesh.tp * expert_mesh.dp == dp * cp * tp
        # With tp=4, cp=2, dp=8: dp * cp * tp = 8 * 2 * 4 = 64
        # But we'll provide expert_mesh that multiplies to 32 (invalid)
        expert_mesh = ParallelConfig.ExpertParallelCfg(
            ep=4,  # 4 * 2 * 4 = 32 ✗ (should be 64)
            tp=2,
            dp=4,
        )

        with pytest.raises(AssertionError):
            ParallelConfig(
                tp=4,
                cp=2,
                pp=2,
                dp=8,
                expert_mesh=expert_mesh,
                vpp=1,
                sp_enabled=True,
                zero_level=ParallelConfig.ZeroLevel.PARTITION_OPTIMIZER,
            )

    def test_zero_level_enum_values(self):
        """Test ZeroLevel enum values."""
        assert ParallelConfig.ZeroLevel.NONE == 0
        assert ParallelConfig.ZeroLevel.PARTITION_OPTIMIZER == 1
        assert ParallelConfig.ZeroLevel.PARTITION_GRADIENTS == 2
        assert ParallelConfig.ZeroLevel.PARTITION_PARAMETERS == 3

    def test_different_world_sizes(self):
        """Test various world size configurations."""
        test_cases = [
            (1, 1, 1, 1, 1),  # Single device
            (2, 1, 1, 1, 2),  # TP only
            (1, 1, 2, 1, 2),  # PP only
            (1, 1, 1, 4, 4),  # DP only
            (2, 2, 2, 2, 16),  # All dimensions
            (8, 1, 4, 16, 512),  # Large cluster
        ]

        for tp, cp, pp, dp, expected_world_size in test_cases:
            config = ParallelConfig(
                tp=tp,
                cp=cp,
                pp=pp,
                dp=dp,
                expert_mesh=None,
                vpp=1,
                sp_enabled=False,
                zero_level=ParallelConfig.ZeroLevel.PARTITION_OPTIMIZER,
            )
            assert config.world_size() == expected_world_size

    def test_expert_mesh_constraint_with_cp_equals_1(self):
        """Test expert mesh constraint when cp=1."""
        # With tp=8, cp=1, dp=64: dp * cp * tp = 64 * 1 * 8 = 512
        expert_mesh = ParallelConfig.ExpertParallelCfg(
            ep=8,  # 8 * 8 * 8 = 512 ✓
            tp=8,
            dp=8,
        )

        config = ParallelConfig(
            tp=8,
            cp=1,
            pp=4,
            dp=64,
            expert_mesh=expert_mesh,
            vpp=1,
            sp_enabled=False,
            zero_level=ParallelConfig.ZeroLevel.PARTITION_OPTIMIZER,
        )

        assert config.expert_mesh.ep == 8
        assert config.expert_mesh.tp == 8
        assert config.expert_mesh.dp == 8
        # Verify the constraint
        assert (
            config.expert_mesh.ep * config.expert_mesh.tp * config.expert_mesh.dp
            == config.dp * config.cp * config.tp
        )

    def test_expert_mesh_with_different_tp_values(self):
        """Test that expert_mesh.tp can differ from main tp."""
        # Main config has tp=4, but expert_mesh has tp=1
        # With tp=4, cp=2, dp=8: dp * cp * tp = 8 * 2 * 4 = 64
        expert_mesh = ParallelConfig.ExpertParallelCfg(
            ep=16,  # 16 * 1 * 4 = 64 ✓
            tp=1,  # Different from main tp=4
            dp=4,
        )

        config = ParallelConfig(
            tp=4,
            cp=2,
            pp=2,
            dp=8,
            expert_mesh=expert_mesh,
            vpp=1,
            sp_enabled=True,
            zero_level=ParallelConfig.ZeroLevel.PARTITION_OPTIMIZER,
        )

        assert config.tp == 4  # Main TP
        assert config.expert_mesh.tp == 1  # Expert TP (different)
        assert config.expert_mesh.ep == 16
        assert config.expert_mesh.dp == 4

    def test_vpp_values(self):
        """Test different VPP (Virtual Pipeline Parallel) values."""
        for vpp in [1, 2, 4, 8]:
            config = ParallelConfig(
                tp=2,
                cp=1,
                pp=4,
                dp=2,
                expert_mesh=None,
                vpp=vpp,
                sp_enabled=False,
                zero_level=ParallelConfig.ZeroLevel.PARTITION_OPTIMIZER,
            )
            assert config.vpp == vpp

    def test_sp_enabled_flag(self):
        """Test sequence parallel enablement flag."""
        config_sp_on = ParallelConfig(
            tp=2,
            cp=1,
            pp=2,
            dp=2,
            expert_mesh=None,
            vpp=1,
            sp_enabled=True,
            zero_level=ParallelConfig.ZeroLevel.PARTITION_OPTIMIZER,
        )
        assert config_sp_on.sp_enabled is True

        config_sp_off = ParallelConfig(
            tp=2,
            cp=1,
            pp=2,
            dp=2,
            expert_mesh=None,
            vpp=1,
            sp_enabled=False,
            zero_level=ParallelConfig.ZeroLevel.PARTITION_OPTIMIZER,
        )
        assert config_sp_off.sp_enabled is False

    def test_edge_case_single_device_with_expert_mesh(self):
        """Test edge case with single device and expert mesh."""
        # With tp=1, cp=1, dp=1: dp * cp * tp = 1 * 1 * 1 = 1
        expert_mesh = ParallelConfig.ExpertParallelCfg(
            ep=1,  # 1 * 1 * 1 = 1 ✓
            tp=1,
            dp=1,
        )

        config = ParallelConfig(
            tp=1,
            cp=1,
            pp=1,
            dp=1,
            expert_mesh=expert_mesh,
            vpp=1,
            sp_enabled=False,
            zero_level=ParallelConfig.ZeroLevel.NONE,
        )

        assert config.world_size() == 1
        assert config.expert_mesh.ep == 1
        assert config.expert_mesh.tp == 1
        assert config.expert_mesh.dp == 1

    def test_large_scale_configuration(self):
        """Test large-scale cluster configuration."""
        # Simulating a large cluster with 12288 GPUs
        # tp=8, cp=4, pp=6, dp=64 => 8 * 4 * 6 * 64 = 12288
        expert_mesh = ParallelConfig.ExpertParallelCfg(
            ep=128,  # 128 * 8 * 2 = 2048 = dp * cp * tp = 64 * 4 * 8
            tp=8,
            dp=2,
        )

        config = ParallelConfig(
            tp=8,
            cp=4,
            pp=6,
            dp=64,
            expert_mesh=expert_mesh,
            vpp=2,
            sp_enabled=True,
            zero_level=ParallelConfig.ZeroLevel.PARTITION_OPTIMIZER,
        )

        assert config.world_size() == 12288
        assert config.expert_mesh.ep == 128
        # Verify constraint
        assert (
            config.expert_mesh.ep * config.expert_mesh.tp * config.expert_mesh.dp
            == config.dp * config.cp * config.tp
        )



import dataclasses
from enum import Enum


class ActivationCheckpointingType(Enum):
    NONE = 1
    SELECTIVE = 2
    SUPER_SELECTIVE = 3
    FULL = 4

    @staticmethod
    def from_str(str_key: str) -> "ActivationCheckpointingType":
        return {
            "none": ActivationCheckpointingType.NONE,
            "selective": ActivationCheckpointingType.SELECTIVE,
            "super-selective": ActivationCheckpointingType.SUPER_SELECTIVE,
            "full": ActivationCheckpointingType.FULL,
        }[str_key]


@dataclasses.dataclass
class CrossDCConfig:
    """Configuration for cross-DC training.

    Attributes:
        n_dcs: Number of data centers in the ring topology
        interconnect_bandwidth_gbps: Total cross-DC bandwidth in Gbps
        interconnect_latency_s: Latency between adjacent DCs in seconds
    """

    n_dcs: int
    interconnect_bandwidth_gbps: float
    interconnect_latency_s: float

    def interconnect_bandwidth_bytes_per_sec(self) -> float:
        """Convert bandwidth from Gbps to bytes per second.

        Returns:
            Bandwidth in bytes per second
        """
        return self.interconnect_bandwidth_gbps * 1e9 / 8  # Gbps to bytes/sec



"""
determine the best instance placement and rank assignment to ensure DP ring
communication can occur with the fewest possible network hops.
"""

from argparse import ArgumentParser

import boto3  # type: ignore[import-untyped, import-not-found]

import dlcalc.utils.cluster.ec2
import dlcalc.utils.cluster.kubernetes
import dlcalc.utils.cluster.topology
from dlcalc.utils.cluster.topology import TreeNode
from dlcalc.utils.math import safe_divide


def main() -> None:
    parser = ArgumentParser(__doc__)
    parser.add_argument(
        "-t",
        "--tp-degree",
        type=int,
        required=True,
        help="tensor parallel (TP) degree",
    )
    parser.add_argument(
        "-p",
        "--pp-degree",
        type=int,
        required=True,
        help="pipeline parallel (PP) degree",
    )
    parser.add_argument(
        "-d",
        "--dp-degree",
        type=int,
        required=True,
        help="data parallel (DP) degree",
    )
    parser.add_argument(
        "-n",
        "--accelerators-per-node",
        type=int,
        default=8,
        help="number of accelerators per node",
    )
    parser.add_argument(
        "-i",
        "--instance-type",
        type=str,
        help="required instance type",
    )
    parser.add_argument(
        "-r",
        "--region",
        type=str,
        help="required AWS region",
    )
    parser.add_argument(
        "-a",
        "--az",
        type=str,
        help="required AWS availability zone",
    )
    args = parser.parse_args()
    tp_degree = args.tp_degree
    pp_degree = args.pp_degree
    dp_degree = args.dp_degree
    accelerators_per_node = args.accelerators_per_node

    n_required_nodes = safe_divide(tp_degree * pp_degree * dp_degree, accelerators_per_node)

    free_instance_ids = dlcalc.utils.cluster.kubernetes.get_free_instances()
    print(f"found {len(free_instance_ids)} free instances")
    instance_type = args.instance_type
    region_name = args.region
    az = args.az

    assert tp_degree <= accelerators_per_node

    # we only care to optimize inter-node DP communication
    # (intra-node DP communication can't be optimized with better instance placement),
    # so we'll ignore the parts of the ring that traverse intra-node links.
    # we're making the assumption that all TP groups are within a single node.
    # since we're looking at things at the instance level, there's only one
    # ring per PP rank.
    dp_ranks_per_node = safe_divide(accelerators_per_node, tp_degree)
    nodes_per_dp_ring = safe_divide(dp_degree, dp_ranks_per_node)
    print(
        f"DP={dp_degree} and {dp_ranks_per_node} DP ranks/node, "
        f"therefore there are {nodes_per_dp_ring} nodes per DP ring"
    )
    node_id_to_node = dlcalc.utils.cluster.topology.build_instance_tree(
        ec2_client=boto3.client("ec2", region_name=region_name),
        accepted_node_instance_types={
            instance_type,
        },
        accepted_node_availability_zones={
            az,
        },
        accepted_instance_ids=free_instance_ids,
    )
    n_discovered_instances = len(node_id_to_node)

    if n_discovered_instances < n_required_nodes:
        raise RuntimeError(
            f"training job requires {n_required_nodes} nodes but found {n_discovered_instances}"
        )
    print(f"discovered {len(node_id_to_node)} instances")

    # find root node
    (root_node,) = (node for node in node_id_to_node.values() if node.parent is None)

    # determine optimal topology
    # we want to select an ordered subset of nodes for optimal adjacency.
    # the most adjacent set of nodes for ring collectives minimizes the number
    # of network nodes that need to be crossed.
    # Physical topology is a tree. in each level we can order by the subtree size,
    # then we'll walk it DFS. this will give us an ordered list of instances that would
    # make relatively suitable rings.
    # TODO. this is sort of a heuristic, will not always find the most optimal
    # placement.
    ordered_nodes = dlcalc.utils.cluster.topology.dfs_tree_leaves_only(root_node)

    if len(ordered_nodes) != n_discovered_instances:
        # TODO. need to debug further to see why not all nodes are reachable by the DFS
        # just warning for now.
        print(
            f"[WARN] expected to traverse {n_discovered_instances} instances "
            f"but only encountered {len(ordered_nodes)}"
        )
        # raise RuntimeError(
        #     f"expected to traverse {len(instance_ids)} nodes "
        #     f"but encountered {len(ordered_nodes)}"
        # )

    ring_assignments: list[list[TreeNode]] = []
    for pp_rank in range(pp_degree):
        ring_assignments.append(ordered_nodes[:nodes_per_dp_ring])
        ordered_nodes = ordered_nodes[nodes_per_dp_ring:]

        print(f"ring for PP={pp_rank}")
        print([n.node_id for n in ring_assignments[pp_rank]])
        traversal_dist = dlcalc.utils.cluster.topology.traversal_distance_of_ring(
            ring_assignments[pp_rank]
        )
        print(
            f"traversal distance: {traversal_dist} -> avg traversal distance {traversal_dist / nodes_per_dp_ring:.2f}"
        )

    # TODO. have this output an MPI rankfile instead
    print("outputting instance list")
    for ring_assignment in ring_assignments:
        for node in ring_assignment:
            print(node.node_id)


if __name__ == "__main__":
    main()



import abc
from abc import ABC
from datetime import timedelta
from enum import Enum

from dlcalc.utils.hardware import MachineSpec
from dlcalc.utils.math import safe_divide


class Dtype(Enum):
    FP32 = "fp32"
    FP16 = "fp16"
    BF16 = "bf16"


class Op(ABC):
    @abc.abstractmethod
    def compute_runtime_s(self, machine_spec: MachineSpec) -> timedelta:
        raise NotImplementedError

    @abc.abstractmethod
    def get_n_params(self, partitioned: bool) -> int:
        raise NotImplementedError


class Transformer(Op):
    pass


class TransformerBlock(Op):
    pass


class Attention(Op):
    pass


class MLP(Op):
    pass


class ExpertMLP(Op):
    def __init__(
        self,
        n_tokens: int,
        n_experts: int,
        hidden_dim: int,
        mlp_hidden_dim: int,
        glu: bool,
        ep: int,
        tp: int,
    ) -> None:
        expert_capacity = None  # TODO.

        self.router = GEMM(
            m=n_tokens,
            k=hidden_dim,
            n=n_experts,
            m_partition_degree=1,  # TODO: proper partitioning
            k_partition_degree=1,
            n_partition_degree=1,
            dtype=Dtype.FP16,  # TODO: proper dtype from config
        )

        self.mlp_up = BatchedGEMM(
            b=n_experts,  # batch size is number of experts
            m=expert_capacity if expert_capacity else 1,  # TODO: proper expert capacity
            k=hidden_dim,
            # following common practice of merging up + gate matmuls in the event
            # we're using GLU.
            n=(mlp_hidden_dim * 2) if glu else mlp_hidden_dim,
        )
        self.mlp_down = BatchedGEMM(
            b=n_experts,  # batch size is number of experts
            m=expert_capacity if expert_capacity else 1,  # TODO: proper expert capacity
            k=(mlp_hidden_dim * 2) if glu else mlp_hidden_dim,
            n=hidden_dim,
        )


class Norm(Op):
    def __init__(self, n_tokens: int, hidden_dim: int, n_tokens_partition_degree: int) -> None:
        self.__n_tokens = n_tokens
        self.__hidden_dim = hidden_dim
        self.__n_tokens_partition_degree = n_tokens_partition_degree

    def compute_runtime_s(self, machine_spec: MachineSpec) -> timedelta:
        # TODO: implement proper normalization runtime calculation
        return timedelta(seconds=0)

    def get_n_params(self, partitioned: bool) -> int:
        # TODO: implement proper parameter count for normalization
        return 0


class SDPA(Op):
    pass


class GEMM(Op):
    def __init__(
        self,
        m: int,
        k: int,
        n: int,
        m_partition_degree: int,
        k_partition_degree: int,
        n_partition_degree: int,
        dtype: Dtype,
    ) -> None:
        self.__m = m
        self.__k = k
        self.__n = n
        self.__m_partition_degree = m_partition_degree
        self.__k_partition_degree = k_partition_degree
        self.__n_partition_degree = n_partition_degree

    def compute_runtime_s(self, machine_spec: MachineSpec) -> timedelta:
        m_partitioned = safe_divide(self.__m, self.__m_partition_degree)
        n_partitioned = safe_divide(self.__n, self.__n_partition_degree)
        k_partitioned = safe_divide(self.__k, self.__k_partition_degree)

        n_flops_partitioned = m_partitioned * n_partitioned * k_partitioned

        n_secs = n_flops_partitioned / machine_spec.device_spec.peak_flops

        return timedelta(seconds=n_secs)

    def get_n_params(self, partitioned: bool) -> int:
        if partitioned:
            k_partitioned = safe_divide(self.__k, self.__k_partition_degree)
            n_partitioned = safe_divide(self.__n, self.__n_partition_degree)
            return k_partitioned * n_partitioned
        else:
            return self.__k * self.__n


class BatchedGEMM(Op):
    def __init__(self, b: int, m: int, k: int, n: int) -> None:
        self.b = b
        self.m = m
        self.k = k
        self.n = n

    def compute_runtime_s(self, machine_spec: MachineSpec) -> timedelta:
        # TODO: implement proper batched GEMM runtime calculation
        return timedelta(seconds=0)

    def get_n_params(self, partitioned: bool) -> int:
        # TODO: implement proper parameter count
        return 0



from .math import ceil_divide, product, safe_divide


class Size:
    """representation of how much space something takes, which is derived from
    what the datatype is, and how many elements of that datatype there are.
    """

    def __init__(self, numel: int, bits_per_element: int) -> None:
        self._numel = numel
        self._bits_per_element = bits_per_element

    def numel(self) -> int:
        return self._numel

    def bits(self) -> int:
        return self._numel * self._bits_per_element

    def bytes(self) -> int:
        return self._numel * safe_divide(self._bits_per_element, 8)

    def __add__(self, other: "Size") -> "Size":
        if self._bits_per_element != other._bits_per_element:
            raise ValueError(
                f"attempted addition between two sizes with different bits per element "
                f"{self._bits_per_element} vs {other._bits_per_element}"
            )

        return Size(numel=self._numel + other._numel, bits_per_element=self._bits_per_element)

    def __mul__(self, multiplicand: int) -> "Size":
        return Size(self._numel * multiplicand, self._bits_per_element)

    def __floordiv__(self, divisor: int) -> "Size":
        return Size(self._numel // divisor, self._bits_per_element)

    def __rmul__(self, multiplicand: int) -> "Size":
        return Size(self._numel * multiplicand, self._bits_per_element)

    def __repr__(self) -> str:
        return (
            f"numel: {self._numel * 1e-9:.3f} B, "
            f"bytes: {self._numel * safe_divide(self._bits_per_element, 8) / (1024**3):.3f} GiB"
        )


class TensorRepr:
    """Representation of a (possibly distributed) Tensor,
    has an unpartitioned shape, along with partitioning axes and degree.
    """

    def __init__(
        self,
        unpartitioned_shape: tuple[int, ...],
        partition_spec: dict[int, int],  # axis -> degree
        bits_per_elt: int,
        enforce_evenly_partitionable: bool = True,
    ) -> None:
        if enforce_evenly_partitionable:
            for partition_dim, partition_degree in partition_spec.items():
                if unpartitioned_shape[partition_dim] % partition_degree != 0:
                    raise RuntimeError(
                        f"dim {partition_dim} of {unpartitioned_shape} not divisible by {partition_degree}"
                    )

        self._unpartitioned_shape = unpartitioned_shape
        self._partition_spec = partition_spec
        self._bits_per_elt = bits_per_elt

    def shape(self, partitioned: bool) -> tuple[int, ...]:
        return self.__get_shape(partitioned=partitioned)

    def numel(self, partitioned: bool) -> int:
        return self.__get_numel(partitioned=partitioned)

    def size(self, partitioned: bool) -> Size:
        return Size(
            numel=self.__get_numel(partitioned=partitioned),
            bits_per_element=self._bits_per_elt,
        )

    def __get_shape(self, partitioned: bool) -> tuple[int, ...]:
        if partitioned:
            shape = list(self._unpartitioned_shape)
            for partition_dim, partition_degree in self._partition_spec.items():
                shape[partition_dim] = safe_divide(shape[partition_dim], partition_degree)
            return tuple(shape)
        else:
            return self._unpartitioned_shape

    def __get_numel(self, partitioned: bool) -> int:
        unpartitioned_numel = product(*self._unpartitioned_shape)
        if partitioned and self._partition_spec:
            total_partitioning_degree = product(*self._partition_spec.values())
            return ceil_divide(unpartitioned_numel, total_partitioning_degree)
        else:
            return unpartitioned_numel



# ANSI color codes
_BOLD = "\033[1m"
_END = "\033[0m"
_GREEN = "\033[92m"
_BLUE = "\033[94m"
_YELLOW = "\033[93m"
_CYAN = "\033[96m"
_RED = "\033[91m"
_MAGENTA = "\033[95m"
_GRAY = "\033[90m"

# Export color codes for external use
__all__ = ["_BOLD", "_END", "_GREEN", "_BLUE", "_YELLOW", "_CYAN", "_RED", "_MAGENTA", "_GRAY"]


def print_bold(text: str) -> None:
    print(_BOLD + text + _END)


def print_kv(key: str, value: str, key_width: int = 25, color: str = "") -> None:
    """Print a key-value pair with nice formatting."""
    formatted_key = f"{key}:".ljust(key_width)
    if color:
        print(f"  {_GRAY}{formatted_key}{_END} {color}{value}{_END}")
    else:
        print(f"  {_GRAY}{formatted_key}{_END} {value}")


def print_metric(name: str, value: str, unit: str = "", highlight: bool = False) -> None:
    """Print a metric with optional highlighting."""
    color = _GREEN if highlight else _YELLOW
    if unit:
        print(f"  {name}: {color}{value}{_END} {_GRAY}{unit}{_END}")
    else:
        print(f"  {name}: {color}{value}{_END}")


def print_section_separator() -> None:
    """Print a light section separator."""
    print(f"{_GRAY}{'·' * 80}{_END}")


def print_success(text: str) -> None:
    """Print success message in green."""
    print(f"{_GREEN}✓ {text}{_END}")


def print_warning(text: str) -> None:
    """Print warning message in yellow."""
    print(f"{_YELLOW}⚠ {text}{_END}")


def print_info(text: str) -> None:
    """Print info message in blue."""
    print(f"{_BLUE}ℹ {text}{_END}")


def format_number(num: float, decimals: int = 2) -> str:
    """Format large numbers with appropriate units."""
    if num >= 1e12:
        return f"{num / 1e12:.{decimals}f}T"
    elif num >= 1e9:
        return f"{num / 1e9:.{decimals}f}B"
    elif num >= 1e6:
        return f"{num / 1e6:.{decimals}f}M"
    elif num >= 1e3:
        return f"{num / 1e3:.{decimals}f}K"
    else:
        return f"{num:.{decimals}f}"


def print_h1_header(section_name: str) -> None:
    print()
    print(f"{_CYAN}{'=' * 80}{_END}")
    print(f"{_BOLD}{_CYAN} {section_name}{_END}")
    print(f"{_CYAN}{'=' * 80}{_END}")


def print_h2_header(section_name: str) -> None:
    print()
    print(f"{_BOLD}{_BLUE}▶ {section_name}{_END}")
    print(f"{_BLUE}{'-' * 60}{_END}")


def get_color_by_intensity(value: float, high_threshold: float, medium_threshold: float) -> str:
    """Get color based on intensity value.

    Args:
        value: The value to evaluate
        high_threshold: Values >= this are considered high (red)
        medium_threshold: Values >= this are considered medium (yellow)

    Returns:
        Color code string (red for high, yellow for medium, green for low)
    """
    if value >= high_threshold:
        return _RED
    elif value >= medium_threshold:
        return _YELLOW
    else:
        return _GREEN


def get_color_by_percentage(percentage: float) -> str:
    """Get color based on percentage value.

    Common thresholds for percentages:
    - >= 30%: Red (dominant)
    - >= 15%: Yellow (significant)
    - < 15%: Green (small)

    Args:
        percentage: The percentage value (0-100)

    Returns:
        Color code string
    """
    return get_color_by_intensity(percentage, 30, 15)


def get_color_by_time_ms(time_ms: float) -> str:
    """Get color based on time in milliseconds.

    Common thresholds for compute time:
    - >= 0.5ms: Red (compute-heavy)
    - >= 0.2ms: Yellow (medium)
    - < 0.2ms: Green (light)

    Args:
        time_ms: Time in milliseconds

    Returns:
        Color code string
    """
    return get_color_by_intensity(time_ms, 0.5, 0.2)


def get_color_for_component_percentage(percentage: float) -> str:
    """Get color for transformer block component based on percentage.

    Uses specialized thresholds for block-level components:
    - >= 20%: Red (dominant within block)
    - >= 10%: Yellow (significant within block)
    - < 10%: Green (small within block)

    Args:
        percentage: The percentage value (0-100)

    Returns:
        Color code string
    """
    return get_color_by_intensity(percentage, 20, 10)
